// Sparse implementation of FreqMNAM using Kundert-Vincentelli package.
// by Carlos E. Christoffersen

#ifndef FreqMNAM_h
#define FreqMNAM_h 1

#include "../network/Circuit.h"
#include <cassert>

extern "C"
{
  #include <spMatrix.h>
}

class FreqMNAM
{
	public:

  FreqMNAM(const ExtDoubleVector& fvec, Circuit* cir, const ElemFlag& mask);
  FreqMNAM(const DenseDoubleVector& fvec, Circuit* cir, const ElemFlag& mask);
  ~FreqMNAM();

  //--------------------------------------------------------------
  //Methods related to the frequency vector
  //--------------------------------------------------------------

  // Get the dimension of frequency vector
  inline unsigned getFVecDim()
	{
		return unsigned(freq_vec.size());
	}

  // Get one frequency from freq. vector.
  inline double getFreq(const unsigned& findex)
	{
		return freq_vec[findex];
	}

  // Get one copy of the frequency vector.
  inline DenseDoubleVector getFreqVec()
	{
		return freq_vec;
	}

  // Method to set the current frequency
  void setFreqIndex(const unsigned& findex);

  // This routine updates the frequency vector and regenerates
  // the matrices at all frequencies but DC.
	void resetFreqVec(const ExtDoubleVector& fvec);
  void resetFreqVec(const DenseDoubleVector& fvec);

  // Get the current frequency
  inline const double& getFreq()
	{
		return freq_vec[current_findex];
	}

  //--------------------------------------------------------------
  // Methods to operate the matrices
  //--------------------------------------------------------------

  // Get matrix dimension
  unsigned getDim();

  // Reorder if necesary and factor the all the matrices
  void factorAll() throw(string&);

  // Reorder if necesary and factor one matrix
  void factor(const unsigned& findex) throw(string&);

  // Get the source vector at a frequency
  void getSource(const unsigned& findex, ExtComplexVector& sourceV);

  // Solve one matrix using the fixed source vector
  void solve(const unsigned& findex, DenseComplexVector& result);

  // Solve one matrix for a given source vector
  void solve(const unsigned& findex, const DenseComplexVector& sourceV,
	DenseComplexVector& result);

	// Solve one matrix for a given source vector
  void solve(const unsigned& findex, const ExtComplexVector& sourceV,
	DenseComplexVector& result);

  //--------------------------------------------------------------
  // Methods to fill the matrices used by the elements
  //--------------------------------------------------------------

  inline void setElement(const unsigned& row, const unsigned& col,
	const double_complex& val)
	{
		// Add error checking
		double* pElement = spGetElement(mnam_vec[current_findex], row, col);
		spADD_COMPLEX_ELEMENT(pElement, val.real(), val.imag());
	}

  inline void setAdmittance(const unsigned& term1, const unsigned& term2,
	const double_complex& val)
	{
		// Add error checking
		spGetAdmittance(mnam_vec[current_findex], term1, term2, &spt);
		spADD_COMPLEX_QUAD(spt, val.real(), val.imag());
	}

  inline void setQuad(const unsigned& row1, const unsigned& row2,
	const unsigned& col1, const unsigned& col2,
	const double_complex& val)
	{
		// Add error checking
		spGetQuad(mnam_vec[current_findex], row1, row2, col1, col2, &spt);
		spADD_COMPLEX_QUAD(spt, val.real(), val.imag());
	}

  inline void setOnes(const unsigned& pos, const unsigned& neg,
	const unsigned& eqn)
	{
		// Add error checking
		spGetOnes(mnam_vec[current_findex], pos, neg, eqn, &spt);
		// No need for macro since the ones are already set.
	}

  //--------------------------------------------------------------
  // Methods to fill the source vectors
  //--------------------------------------------------------------

  // Set one element of the current source vector
  inline void setSource(const unsigned& row, const double_complex& val)
  {
    assert(row <= dimension);
    if (row)
      sourceV_vec[current_findex][row-1] = val;
  }

  // Add value to a pair of elements of the current source vector
  void addToSource(const unsigned& pos, const unsigned& neg,
	const double_complex& val);

  //--------------------------------------------------------------
  // Methods for debugging
  //--------------------------------------------------------------

  void print(int i = -1);

	private:

  // utility function used to fill the matrices
  inline void addElem(const unsigned& i, const unsigned& j,
	DoubleSparseColMatrix*& M1, double val)
	{
		// This is needed because we do not know if the element exists
		// in the matrix.
		double currval((*M1)(i - 1, j - 1));
		(*M1)(i - 1, j - 1) = currval + val;
	}

  // The frequency vector
  DenseDoubleVector freq_vec;

  // The current frequency index
  unsigned current_findex;

  // Dimension of the system
  unsigned dimension;

  // Mask to select elements to be included in matrix
  ElemFlag mymask;

  // The circuit from which to extract the matrices.
  Circuit* my_circuit;

  // Use an element vector to hold the selected elements and
  // save time each time an update is requested.
  ElementVector elem_vec;

  // The vector that holds the matrix pointers
  char** mnam_vec;

  // The source vectors
  double_complex** sourceV_vec;

  // Template used by sparse package
  struct spTemplate spt;

};

#endif

