#include "SVTran3.h"
#include "TimeMNAM.h"
#include "TimeDomainSV.h"
#include "NLSInterface.h"
#include "Euler.h"
#include "Trapezoidal.h"
#include <iostream>

extern int superLU_print_enable ;

extern "C"
{
	#include "../inout/ftvec.h"
	#include "../inout/report.h"
}

void buildTIncidence(ElemFlag mask, Circuit*& my_circuit,
IntDenseColMatrix& T, ElementVector& elem_vec,
int& n_states, int& max_n_states);

// Static members
const unsigned SVTran3::n_par = 15;
FILE *pFile;
// Element information
ItemInfo SVTran3::ainfo =
{
  "SVTran3",
  "State-Variable-Based Time-Marching Transient Analysis with Newton Iterations",
  "Sonali Luniya",
  DEFAULT_ADDRESS
};

// Parameter information
ParmInfo SVTran3::pinfo[] =
{
  {"tstop", "Stop time (s)", TR_DOUBLE, true},
  {"tstep", "Maximum time step (s)", TR_DOUBLE, true},
  {"nst", "No save time (s)", TR_DOUBLE, false},
  {"deriv", "Approximate derivatives or use automatic diff.", TR_INT, false},
  {"msv", "Use Msv flag", TR_BOOLEAN, false},
  {"im", "Integration method", TR_INT, false},
  {"savenode", "Save node voltages", TR_BOOLEAN, false},
  {"permc_spec", "Permutation ordering to factor Msv (0, 1 or 2)",
	TR_INT, false},
  {"out_steps", "Number of steps skipped for output simulation progress",
	TR_INT, false},
  {"gcomp", "Compensation network conductance (S)", TR_DOUBLE, false},
  {"cm" , "Compensation method ",TR_DOUBLE, false},
  {"reltol", "Relative Error" , TR_DOUBLE, false},
  {"abstol", "Absolute Error" , TR_DOUBLE, false},
  {"max_iter", "Maximum number of iterations", TR_INT,false},
  {"tmin" , "Minimum time step", TR_DOUBLE , false}
};


SVTran3::SVTran3() : Analysis(&ainfo, pinfo, n_par), ls_size(0),
superLU(false)
{
  // Parameter stuff
  paramvalue[0] = &(tf);
  paramvalue[1] = &(h);
  paramvalue[2] = &(nst = zero);
  paramvalue[3] = &(deriv = 0);
  paramvalue[4] = &(use_msv = true);
  paramvalue[5] = &(int_method = 0);
  paramvalue[6] = &(savenode = true);
  paramvalue[7] = &(permc_spec = 2);
  paramvalue[8] = &(out_steps = 100);
  paramvalue[9] = &(gcomp = 1e-4);
  paramvalue[10] = &(cm=1.0);
  paramvalue[11] = &(reltol=0.01);
  paramvalue[12] = &(abstol=0.0001);
  paramvalue[13] = &(max_iter=5000);
  paramvalue[14] = &(tmin = 1e-16);
}


SVTran3::~SVTran3()
{
  // assert(!superLU);
}

void SVTran3::run(Circuit* cir)
{
  nt=0;
  total_iteration=0;
  scount=0;
  col_count=0;
  mnam_scount=0;
  cmflag=0;
  sparsity=0;
  cktorder=1;
  prev_timestep=0;
  new_timestep=0;
  ptime=0;
  net_timestep=0;

  net_timestep=h;
  if(cm)
	{
		dummy_run(cir);
		superLU=false;
		dummy_run(cir);
		cmflag=0;
  }
  else
	{
    cmflag=1;
    dummy_run(cir);
  }
}

void SVTran3::dummy_run(Circuit* cir)
{
	  changestep=0;
  this->cir = cir;
  char msg[80];
  superLU_print_enable = 0;
  double error = zero, maxerror = zero, maxerror1=zero;
  int iteration =0,itype=0, flag =0;
  ptime=0;
  n_states = 0;
  int max_n_states = 0;
  // h=net_timestep;
  h=net_timestep/4.0;
  firsttime=2;
  double ctime = zero;

  report(MESSAGE, "*** State Variable-Based Transient Analysis using Sparse matrices and Newton Iterations with NEW Time step control***");
  sepLine();
  newLine();

  // Build time domain MNAM
  ElemFlag mnam_mask(LINEAR);
  TimeMNAM mnam(cir, mnam_mask);
  // Build T matrix and nonlinear element vector


  ElemFlag mask(NONLINEAR);
  buildTIncidence(mask, cir, T, elem_vec, n_states, max_n_states);
  if(cm){
    if(!cmflag){
      rcomp = new double[n_states];
      for(int i=0; i <n_states; i++){
	if(gcomp)
	  rcomp[i]=1.0/gcomp;}
    }
  }
  else{
    rcomp = new double[n_states];
    for(int i=0; i <n_states; i++){
      if(gcomp)
	rcomp[i]=1.0/gcomp;}
  }

  // Add the compensation resistors to the MNAM
  for (int i = 0; i < n_states; i++) {
    if(rcomp[i])
      mnam.setMAdmittance(T(0,i), T(1,i) , 1.0/rcomp[i]);
  }

  // Go through all the nonlinear elements
  int n_elem = elem_vec.size();
  cns.resize(n_elem);
  nss.resize(n_elem);
  n_sec_states = 0;
  for (int k=0; k < n_elem; k++) {
    cns[k] = elem_vec[k]->getNumberOfStates();
    nss[k] = elem_vec[k]->getNumberOfSecStates();
    n_sec_states += nss[k];
  }


  // Setup simulation variables (use circular vectors)
  ls_size = mnam.getDim();
  list_size = int(((tf - nst) / h + 3)*2);
  if (savenode)
    cU = new CircVector(list_size, ls_size);
  else
    cU = new CircVector(51, ls_size);
  if(cm){
    if(!cmflag){ //initialize only once
      cX = new CircVector(list_size, n_states);
      cY = new CircVector(list_size, n_sec_states);
    }
  }
  else{
    cX = new CircVector(list_size, n_states);
    cY = new CircVector(list_size, n_sec_states);
  }
  cVnl = new CircVector(list_size, n_states);
  cInl = new CircVector(list_size, n_states);
  cTime = new CircVector(list_size, 1);
  cTimestep = new CircVector(list_size,1);
  ssv = new double[ls_size+n_states];
  rhs_vec= new double[ls_size+n_states];
  sf = new double [ls_size];

  DoubleDenseColMatrix Ju(n_states , n_states);
  DoubleDenseColMatrix Ji(n_states , n_states);
  DoubleDenseColMatrix rhs_temp(ls_size, n_states);

  DenseIntVector myindex(ls_size,0);

  //Set initial conditions
  LIntegMethod *l_im;
  LIntegMethod *l_im2;
  NLIntegMethod *nl_im;
  NLIntegMethod *nl_im2;
  if (int_method == 1) {
    l_im2 = new LTrapezoidal(cU, &mnam);
    nl_im2 = new NLTrapezoidal(cX,cY,cU,cTime, h);
    l_im = new LEuler(cU, &mnam);
    nl_im = new NLEuler(cX,cY,cU,cTime,h);
  }
  else {
    l_im2 = new LEuler(cU, &mnam);
    nl_im2 = new NLEuler(cX,cY,cU,cTime,h);
    l_im = new LTrapezoidal(cU, &mnam);
    nl_im = new NLTrapezoidal(cX,cY,cU,cTime,h);
  }
  DoubleSparseColMatrix M1(ls_size, ls_size);
  DoubleSparseColMatrix M2(ls_size, ls_size);
  DoubleSparseColMatrix *M;
  l_im->buildMd(M1, h);
  l_im2->buildMd(M2,h);

  // Create the interface class
  // Begin main loop: fixed time step by now
  //double residual = zero;
  report(MESSAGE, "--- Starting transient simulation ...\n");
  sepLine();
  report(MESSAGE, "| Step \t| Time (s) \t|");
  sepLine();

  //Adding MNAM
  double tx =0;
  rhs_Jv_X = new double[n_states];
  elem_val = new double[(ls_size+n_states)*(ls_size+n_states)];
  row_index = new int[(ls_size+n_states)*(ls_size+n_states)];
  col_pointer = new int[ls_size+n_states+1];
  int myT[n_states][ls_size], tmpidx=0;
  for(int i=0 ; i <n_states;i ++)
    for(int j=0; j<ls_size; j++)
      myT[i][j]=0;
  for(int k =0 ; k < n_states ; k++ ){	//Adding T
    if(T(0,k)==T(1,k)){
      T(0,k) =0;
      T(1,k)=0;
    }
    if (T(0,k)){
      tmpidx =T(0,k) -1 ;
      myT[k][tmpidx] = 1;
      myindex[tmpidx]=1;
    }
    if(T(1,k) ){
      tmpidx =T(1,k) -1 ;
      myT[k][tmpidx] =-1;
      myindex[tmpidx]=1;
    }
  }

  m_i = M1.begin();   //Adding MNAM
  iend = M1.end();
  for (; m_i != iend; ++m_i) {
    col_pointer[col_count++] = scount;
    m_j = (*m_i).begin(); jend = (*m_i).end();
    for (; m_j != jend; ++m_j) {
      elem_val[scount] = (*m_j);
      row_index[scount]=m_j.row();
      scount++;

    }

    for(int i=0;i<n_states ;i++){ //Adding T
      if(myT[i][col_count-1]){
	elem_val[scount] = myT[i][col_count-1];
	row_index[scount]=ls_size+i;
	scount++;
      }
    }
  }
  int tcount=0,tcol=0;
  double *temp_mnam1, *temp_mnam2,*temp_mnam;
  temp_mnam1 = new double[scount];
  temp_mnam2 = new double[scount];
  temp_mnam = new double[scount];

  m_i = M2.begin();
  iend = M2.end();
  for (; m_i != iend; ++m_i) {
    tcol++;
    m_j = (*m_i).begin(); jend = (*m_i).end();
    for (; m_j != jend; ++m_j) {
      temp_mnam2[tcount] = (*m_j);
      tcount++;
    }
    for(int i=0;i<n_states ;i++){ //Adding T
      if(myT[i][tcol-1]){
	temp_mnam2[tcount] = myT[i][tcol-1];
	tcount++;
      }
    }
  }


  memcpy(temp_mnam1, elem_val, sizeof(double)*scount);
  mnam_scount = scount;
  double FE_predX[n_states];
  double FE_predU[ls_size];
  Integ_predX = new double[n_states];
  Integ_predU = new double[ls_size];
  Integ_predVnl = new double[n_states];

  n_tsteps = int(tf / h + 1);
  tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),
			  &(cVnl->getCurrent()[0]),
			  &(cInl->getCurrent()[0]),max_n_states,nl_im);
  if(!cmflag)
	{
    mtl::set_value(cX ->getCurrent(), zero);
    cX ->advance();
    mtl::set_value(cX ->getCurrent(), zero);
    cX -> advance();
    mtl::set_value(cY->getCurrent(), zero);
    cY ->advance();
    mtl::set_value(cY ->getCurrent(), zero);
    cY -> advance();
  }
  cTime ->getCurrent()[0]=-h;
  cTime->advance();
  cTime ->getCurrent()[0]=zero;
  cTime ->advance();

  cTimestep->getCurrent()[0]=h;
  cTimestep->advance();
  ctime=h;
  prev_timestep=h;
  while(ptime<tf)
	{
    for( itype=0 ; itype<2; itype++){
//      cout<<"Current time: "<<ctime <<" itype "<<itype<<" Time step: "<<h<<endl;
      iteration = 0;

      if(itype==1){
	if(changestep){
	  nl_im2->changeStep(h);
	  l_im2->buildMd(M1,h);}
	else
	  memcpy(temp_mnam,temp_mnam2,sizeof(double)*mnam_scount);
      }
      else {
	if(changestep){
	  nl_im->changeStep(h);
	  l_im->buildMd(M1,h); }
	else
	  memcpy(temp_mnam,temp_mnam1,sizeof(double)*mnam_scount);
      }
      if(changestep){
	m_i = M1.begin();   //Adding MNAM
	iend = M1.end();
	tcol=0;
	tcount=0;
	for (; m_i != iend; ++m_i) //col index
	  {
	    tcol++;
	    m_j=(*m_i).begin(); jend = (*m_i).end();
	    for (; m_j != jend; ++m_j) {
	      temp_mnam[tcount] =(*m_j) ;
	      tcount++;
	    }
	    for(int i=0;i<n_states ;i++){ //Adding T
	      if(myT[i][tcol-1]){
		temp_mnam[tcount] = myT[i][tcol-1];
		tcount++;
	      }
	    }
	  }
      }
      cTime->getCurrent()[0] = ctime;
      cTimestep->getCurrent()[0] = h;
      tdsv->setTime(&(cX->getCurrent()[0]),&(cVnl->getCurrent()[0]),
		    &(cInl->getCurrent()[0]),nt, ctime, h);//first_time=true
      nl_im->predictX(n_states,FE_predX,cTimestep);
      for(int i=0;i<n_states;i++)
	cX->getCurrent()[i] = FE_predX[i];
      DenseDoubleVector& Inl = cInl->getCurrent();
      DenseDoubleVector& Xnl = cX->getCurrent();
      DenseDoubleVector& Vnl = cVnl->getCurrent();
      DenseDoubleVector& Ul  = cU->getCurrent();
      // Update compressed source vector
      memset( sf,0,sizeof(double)*ls_size);
      if(itype==1)
	l_im2->buildSf(sf, ctime);
      else
	l_im->buildSf(sf,ctime);

      /*for(int i =0; i <n_states;i++)
	cout<<"X[i]"<<cX->getCurrent()[i]<<endl;
	sepLine();*/
      do  //Start Newton Iterations
	{
	  scount=mnam_scount;
	  // for(int i=0;i<ls_size;i++)
	  //ssv[i]=sf[i];
	  memcpy(ssv,sf,sizeof(double)*ls_size);
	  memset(elem_val,0,sizeof(double)*((ls_size+n_states)*(ls_size+n_states)));
	  memcpy(elem_val,temp_mnam,sizeof(double)*(mnam_scount));
	  col_count = ls_size;
	  mtl::set_value(rhs_temp, 0);
	  int ibase =0,jbase=0 ;
	  memset(rhs_Jv_X, 0 , n_states*sizeof(double));
	  int n_elem = elem_vec.size();
	  int prev_scount=0;

	  for (int k = 0; k < n_elem; k++)
	    {
	      int c_cns = elem_vec[k]->getNumberOfStates();
	      // Set base index in interface object
	      tdsv->setBase(ibase,cns[k],jbase, nss[k]);
	      //elem_vec[k]->deriv_svTran(tdsv);
	      elem_vec[k]->svTran(tdsv);
	      //tdsv->clearflag();
	      elem_vec[k]->deriv_svTran(tdsv);     // Call element evaluation

	      DoubleDenseColMatrix& Ju_elem = tdsv->getJu();
	      DoubleDenseColMatrix& Ji_elem = tdsv->getJi();

	      for(int m=0 ; m<c_cns; m++){ //Ji col counter
		col_pointer[col_count++]= scount;
		prev_scount = scount;
		for (int j=0; j < c_cns; j++) { //T' row counter
		  int tmpidx=0;
		  if(T(0,j+ibase)&&Ji_elem(j,m))         //multiply with 1
		    {
		      // double tx = Ji_elem(j,m);
		      tmpidx = T(0,j+ibase)-1;
		      int i =0;
		      for( i = prev_scount; i< scount;i++)
			{
			  if(row_index[i]==tmpidx){ //element already exists
			    elem_val[i]+=Ji_elem(j,m);
			    break;}
			  if(row_index[i]>tmpidx)
			    {
			      for(int ti = scount ; ti > i ; ti--)
				{
				  elem_val[ti]= elem_val[ti-1];
				  row_index[ti]=row_index[ti-1];
				}
			      elem_val[i]= Ji_elem(j,m);
			      row_index[i]=tmpidx;
			      scount++;
			      break;
			    }
			}
		      if(i==scount) //add new elem at poistion i
			{
			  elem_val[i]= Ji_elem(j,m);
			  row_index[i]=tmpidx;
			  scount++;
			}
		    }
		  if(T(1,j+ibase)&&Ji_elem(j,m))         //multiply with -1
		    {
		      tmpidx = T(1,j+ibase)-1;
		      int i =0;
		      for( i = prev_scount; i< scount;i++)
			{
			  if(row_index[i]==tmpidx){//element already exists
			    elem_val[i]-=Ji_elem(j,m);
			    break;}
			  if(row_index[i]>tmpidx)
			    {
			      for(int ti = scount ; ti > i ; ti--)
				{
				  elem_val[ti]= elem_val[ti-1];
				  row_index[ti]=row_index[ti-1];
				}
			      elem_val[i]= -Ji_elem(j,m);
			      row_index[i]=tmpidx;
			      scount++;
			      break;
			    }
			}
		      if(i==scount) //add new elem at poistion i
			{
			  elem_val[i]= -Ji_elem(j,m);
			  row_index[i]=tmpidx;
			  scount++;
			}
		    }
		}
		for(int i=prev_scount; i <scount;i++)
		  rhs_temp(row_index[i],m+ibase)=elem_val[i];
		for(int i=0; i < c_cns;i++){
		  elem_val[scount]= -Ju_elem(i,m);
		  row_index[scount]=ls_size+i+ibase;
		  scount++;
		  rhs_Jv_X[i+ibase]+=Ju_elem(i,m)*cX->getCurrent()[m+ibase];
		}
	      }
	      ibase += cns[k];
	      jbase += nss[k];
	    }

	  tdsv->clearflag();

	  //report(MESSAGE, "--Multiplying T'JiXi... \n");
	  multiply(rhs_temp , Xnl, ssv);
	  for(int k=0 ; k <n_states; k++) //Multiplying T'*Inl
	    {
	      if(T(0,k)){
		tmpidx=(T(0,k)) -1 ;
		if(rcomp[k])
		  ssv[tmpidx] =ssv[tmpidx]-(Inl[k]-Vnl[k]/rcomp[k]);
		else
		  ssv[tmpidx] =ssv[tmpidx]-Inl[k];
	      }
	      if(T(1,k) ){
		tmpidx =(T(1,k)) -1;
		if(rcomp[k])
		  ssv[tmpidx] =ssv[tmpidx]+(Inl[k]-Vnl[k]/rcomp[k]);
		else
		  ssv[tmpidx]=ssv[tmpidx]+(Inl[k]);
	      }

	    }

	  for(int i = 0 ; i < n_states ; i++)
	    ssv[ls_size+i] = Vnl[i] - rhs_Jv_X[i];
	  rhs_p = ssv;

	  /* report(MESSAGE,"Printing LHS...\n");
	     for(int i= 0; i < scount; i++){
	     sprintf(msg, "M[%d]= %f, %d ", i , elem_val[i],row_index[i]);
	     report(MESSAGE,msg);
	     }
	     sepLine();
	     report(MESSAGE,"Printing RHS...\n");
	     for(int i= 0; i < ls_size+n_states; i++){
	     sprintf(msg, "RHS[%d]= %f  ", i , ssv[i]);
	     report(MESSAGE,msg);
	     }
	     sepLine();
	  */

	  superLUFactor(M,ssv);
	  // for(int i=0;i<ls_size+n_states;i++)
	  //cout<<"ssv["<<i<<"] = "<<ssv[i]<<endl;
	  error = zero;
	  flag =0 ;
	  for(int j= 0 ; j < ls_size &&flag==0 ; j++){
	    error = abs(Ul[j] - ssv[j] );
	    if(error > abstol)
	      flag = 1;
	  }


	  // for(int j =0 ; j < n_states && flag==0; j++){
	  //error = abs(cX->getCurrent()[j] - ssv[ls_size+j] );
	  //if(error > abstol)
	  // flag = 1; //continue Newton iteration
	  //}

	  if(!cmflag){
	    iteration = max_iter;
	    ptime=tf;
	    flag=1;
	  }
	  iteration++;
	  updateVInl(&(ssv[0]));
	  if(firsttime&&flag==1&&iteration<max_iter)
	    freeSuperLU();

	}while(iteration < max_iter && flag ==1 );

      total_iteration+=iteration;
      // cout<<" # of iterations "<<iteration<<endl;
      int jbase =0,ibase=0;
      int n_elem = elem_vec.size();
      for (int k = 0; k < n_elem; k++)
	{
	  int c_cns = elem_vec[k]->getNumberOfStates();
	  // Set base index in interface object
	  tdsv->setBase(ibase, cns[k],jbase,nss[k]);       // Call element evaluation
	  elem_vec[k]->svTran(tdsv);
	  if(!cmflag){
	    elem_vec[k]->nx=0;
	    elem_vec[k]->ndx=0;}
	   ibase += cns[k];
	   jbase+= nss[k];
	}
      if(flag==1)
	break;
      if(itype==0)
	{
	  for(int i=0; i <n_states; i++)
	    {
	      Integ_predX[i]=cX->getCurrent()[i];
	      Integ_predVnl[i]= cVnl->getCurrent()[i];
	      //cX->getCurrent()[i] =cX->getPrevious(1)[i];
	    }
	  for(int i =0; i < ls_size; i++)
	    Integ_predU[i] = cU->getCurrent()[i];
	  delete tdsv;
	  tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),&(cVnl->getCurrent()[0]),
				  &(cInl->getCurrent()[0]),max_n_states,nl_im2);
	  tdsv->clearflag();
	}
    }//end of for (itype ) loop
    prev_timestep=h;
    new_timestep= h;
    changestep=0;
    if(iteration >= max_iter){
      sprintf(msg , "At time step %g Newton Raphson  didn't converge \t",ctime);
      report(MESSAGE, msg);
      // if(itype==1){
	new_timestep = h = h/2.0;
	changestep=1;
	//}
    }
    else{
      if(!firsttime){
	PredictNewStep(ssv,&new_timestep);
	// h=new_timestep;

	if(new_timestep>0.85 *prev_timestep){
	  new_timestep=h;
	  cktorder=2;
	  PredictNewStep(ssv,&new_timestep);
	  if(new_timestep<=1.05*prev_timestep)
	    cktorder=1;
	  //accept this iteration
	     h= new_timestep;
	     changestep=2;
	}
	else{   // reject this iteration
	  changestep=1;
	  h = new_timestep;}
	if(h <= tmin ) {
	  if(prev_timestep > tmin)
	    h = tmin;
	  else{
	    cout<<"Timestep too small. Try increasing tmin and reltol.."<<endl;
	    exit(1);
	  }
	}
	if(h>net_timestep)
	  h=net_timestep;
      }
      if(firsttime)
	firsttime--;
    }
    if(cm && (!cmflag)){ //Find rcomp
      //report(MESSAGE,"Printing Ii\n");
      //for(int i = 0; i < n_states; i ++)
      //cout<<"I["<<i<<"] = "<<cInl->getCurrent()[i]<<"X["<<i<<"] "<<cX->getCurrent()[i]<<endl;
      DenseDoubleVector tmp_vec4(ls_size,0);
      DenseDoubleVector linear_i(ls_size,0);
      DenseDoubleVector delta_i(ls_size,0);
      //multiply T' and Inl
      for(int k=0 ; k <n_states; k++){
	rcomp[k]=0;
	if(T(0,k)){
	  int tmpidx=(T(0,k)) -1 ;
	  tmp_vec4[tmpidx] =tmp_vec4[tmpidx]+cInl->getCurrent()[k];
	}
	if(T(1,k) ){
	  int tmpidx =(T(1,k)) -1;
	  tmp_vec4[tmpidx] =tmp_vec4[tmpidx]-cInl->getCurrent()[k];
	}

      }

      DenseDoubleVector icomp(n_states,0);
      for(int i=0; i <ls_size; i++){
	//go row by row to compensate delta
	if(myindex[i])
	  {
	    double denom =0;
	    //multiply M1 and U
	    for(int j =0; j <ls_size;j++)
	      {
		double tx = M1(i,j);
		linear_i[i] +=tx * cU->getCurrent()[j];
	      }
	    delta_i[i]= linear_i[i]+tmp_vec4[i];
	    DenseIntVector sum_index(n_states,0);
	    for(int k=0; k <n_states ;k++)
	      { //check which currents are involved
		double tx = myT[k][i];
		if(tx && !icomp[k]){
		  denom += tx*cInl->getCurrent()[k];
		  sum_index[k]++;
		}
	      }
	    for(int k=0; k<n_states; k++)
	      {
		if(sum_index[k])
		  {
		    if((delta_i[i]==0)||(cInl->getCurrent()[k]==0))
		      rcomp[k]=0;
		    else
		      rcomp[k]=(cVnl->getCurrent()[k]*denom)/(delta_i[i]*cInl->getCurrent()[k]);
		    rcomp[k]= abs(rcomp[k]);
		    if(rcomp[k] && (rcomp[k]<100))
		      rcomp[k]=100;
		    // cout<<"rcomp["<<k<<"] "<<rcomp[k]<<endl;
		    icomp[k]=1;
		  }
	      }

	  }


      }
    } //end of compensation calculations


    // Update previous derivatives in integration method (if required)
    if((changestep==2 || changestep==0) && !firsttime){ //going forward
      nl_im2->store();
      l_im2->store();
      for(int i =0 ; i < n_states ; i++)
	cX->getCurrent()[i]= Integ_predX[i]; //done only for storing the previous deriv
      nl_im->store();
      l_im->store();

    }

    delete tdsv;

    // Advance the circular lists for the next step.
    //After dummy run ,run is started with the values of cX from dummy run
    //as Adolc does not erase the tape. ????

    if(cmflag &&  (changestep==2)||(changestep==0)) { //going forward
      nt++;
      cU->advance();
      cX->advance();
      cY->advance();
      cVnl->advance();
      cInl->advance();
      cTime->advance();
      cTimestep->advance();
    }
    //set initial cX at next time step = calculated cX at this time step

    if((changestep==1)&&cmflag)   //going backward , rejection
      cX->getCurrent()= cX->getPrevious(1);
    else{
      for(int i =0 ; i < n_states ; i++)
	cX->getCurrent()[i]=ssv[ls_size+i] ;  //going forward
      //fprintf(pFile, " %f \n",Integ_predU[4]);

    }
    if(nt>=0)  {  //transient constructor
      tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),&(cVnl->getCurrent()[0]),
			      &(cInl->getCurrent()[0]),max_n_states,nl_im);
      tdsv->clearflag();
    }
    else {          //dc constructor
      tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),
			      &(cVnl->getCurrent()[0]),
			      &(cInl->getCurrent()[0]),max_n_states);
      tdsv->clearflag();
    }

    // Print table line
    if (!(nt % out_steps)) {
      sprintf(msg, "| %g  \t| %g \t|", h, ctime);
      report(MESSAGE, msg);
    }
    if(cmflag){
      ptime=cTime->getPrevious(1)[0];
      ctime = ptime + h;
    }


  } // end time loop

  superLU = true;
  if(cmflag)
    doOutput();
  cmflag=1;
  delete nl_im;
  delete l_im;
  //freeSuperLU();
  delete cInl;
  delete cVnl;
  //delete cX;
  delete cU;
  return;
}


void SVTran3::superLUFactor(DoubleSparseColMatrix*& M, double* ssv)
{
  // Call freeSuperLU to release memory
  //  assert(!superLU);
  int info,nrhs=1;
  col_pointer[col_count] = scount;
  fact      = 'N';
  equed     = 'N';
  trans     = 'N';
  refact    = 'N';
  panel_size=10;
  relax=5;
  if(firsttime)
    {
      dCreate_CompCol_Matrix(&A, n_states+ls_size, n_states+ls_size, scount,
			     elem_val, // elem_val pointer
			     row_index, // row_index pointer
			     col_pointer, // col_pointer pointer
			     NC, _D, GE);
      //dCreate_Dense_Matrix(&B, n_states+ls_size, nrhs, rhs_p, n_states+ls_size, DN, _D, GE);
      dCreate_Dense_Matrix(&X, n_states+ls_size, nrhs, rhs_p, n_states+ls_size, DN, _D, GE);
      StatInit(panel_size,relax);
      perm_r = new int[n_states+ls_size];
      perm_c = new int[n_states+ls_size];
      etree = new int[ls_size+n_states];
      get_perm_c(permc_spec, &A, perm_c);
      sp_preorder(&refact,&A,perm_c,etree,&AC);
      dgstrf(&refact,&AC,1.0,0.0, relax,panel_size,etree,NULL,0,perm_r,perm_c,&L,&U,&info);
      dgstrs(&trans,&L,&U,perm_r,perm_c,&X,&info);

    }
  else{
     refact='Y';
    dgstrf(&refact,&AC,1.0,0.0, relax,panel_size,etree,NULL,0,perm_r,perm_c,&L,&U,&info);
    dgstrs(&trans,&L,&U,perm_r,perm_c,&X,&info);
  }
  /*
   * Get column permutation vector perm_c[], according to permc_spec:
   *   permc_spec = 0: use the natural ordering
   *   permc_spec = 1: use minimum degree ordering on structure of A'*A
   *   permc_spec = 2: use minimum degree ordering on structure of A'+A
   */
   if (info) {
     report(FATAL, "There was a problem factoring the MNAM.");
     report(MESSAGE, "Try increasing tstep..");
  }
   // NCformat *Ustore;
   //SCformat *Lstore;
   //Lstore = (SCformat *) L.Store;
   //Ustore = (NCformat *) U.Store;
   //report(MESSAGE, msg);

   fflush(stdout);

   // This is not used anymore
   // Destroy_SuperMatrix_Store(&X);
   superLU = true;
   sparsity =(double) scount/((n_states+ls_size)*(n_states+ls_size));
   sparsity = sparsity*100;
   scount=0;
   col_count =0;
}

void SVTran3::freeSuperLU()
{
  assert(superLU);
  // Free vectors
  delete [] etree;
  delete [] perm_c;
  delete [] perm_r;
  // Free supermatrix structures
  Destroy_SuperMatrix_Store(&A);
  Destroy_SuperMatrix_Store(&B);
  // Free L and U matrices
  if ( lwork >= 0 ) {
    Destroy_SuperNode_Matrix(&L);
    Destroy_CompCol_Matrix(&U);
  }
  superLU = false;
}

void SVTran3::multiply(DoubleDenseColMatrix& a ,
DenseDoubleVector& b , double* c)
{
  int imax = a.nrows();
  int jmax = a.ncols();
  for(int i =0 ; i < imax ; i++)
  {
    for(int j=0; j < jmax; j++)
    {
	    c[i] += a(i,j) * b[j];
    }
  }
}


void SVTran3::updateVInl(double* x_p)
{
  memcpy(&(cU->getCurrent()[0]),x_p,sizeof(double)*ls_size);
  memcpy(&(cX->getCurrent()[0]),&(x_p[ls_size]),sizeof(double)*n_states);
}


void SVTran3::PredictNewStep(double* ssv, double *timestep)
{
  double tol=0, diff=0, tmp=0, timetemp=1.7e+18;
//  return;
  switch(cktorder){
  case 1:
    for(int i=0; i < ls_size; i++){
      tol = MAX(fabs(Integ_predU[i]),fabs(ssv[i]))*reltol + abstol;
      diff = (ssv[i] -Integ_predU[i]);
      if(fabs(diff)>=reltol){
	//tmp=  diff/reltol;
	tmp = 6.0 * tol * 2.0 /diff;
	tmp = cTimestep->getCurrent()[0] * sqrt(fabs(tmp));
	//tmp= cTimestep->getCurrent()[0]/2.0;
	timetemp = MIN (timetemp,tmp);
      }
    }
    break;
  case 2:
    for(int i=0; i < ls_size; i++){
      tol = MAX(fabs(Integ_predU[i]),fabs(ssv[i]))*reltol + abstol;
      diff = ssv[i] -Integ_predU[i];
      if(fabs(diff)>=reltol){
	tmp = cTimestep->getCurrent()[0]*6.0*tol*2.0*
	  (cTimestep->getCurrent()[0]+cTimestep->getPrevious(1)[0])/diff;
	tmp =sqrt(fabs(tmp));
	timetemp = MIN (timetemp,tmp);
      }
    }
    break;
  default:
    break;
  }
  *timestep = MIN(2 * *timestep, timetemp);
}


void SVTran3::doOutput()
{
  // First check if the result matrices contain any data
  assert(cX);
  cout<<"Number of state variables :"<<n_states<<endl;
  cout<<"Size of sparse matrix :"<<n_states+ls_size <<"*"<<n_states+ls_size<<endl;
  cout<<"Sparsity of the matrix :"<< sparsity<<"%"<<endl;
  cout<<"Total Number of Newton Iterations :"<<total_iteration<<endl;
  report(MESSAGE, "--- Writing output vectors ...");

  int out_size;
  if(nt>list_size)
    out_size=list_size;
  else
    out_size=nt;

  // Allocate and copy global time vector
  allocTimeV_P(out_size);
  for (int tindex = 0; tindex < out_size; tindex++)
    TimeV_P[tindex] =cTime->getPrevious(out_size-tindex)[0];

  // Temporary vectors
  DenseDoubleVector tmp_x(out_size);
  DenseDoubleVector tmp_i(out_size);
  DenseDoubleVector tmp_u(out_size);

  // Now fill currents and port voltages of time domain devices.
  int current_ns;
  int n_elem = elem_vec.size();   // Number of elements
  int i=0;
  for (int k=0; k < n_elem; k++) {

    current_ns = elem_vec[k]->getNumberOfStates();
    for(int j=0; j< current_ns; j++) {

      // For the current, decompensate while copying.
      for (int tindex=0; tindex < out_size; tindex++) {
				tmp_x[tindex] = cX->getPrevious(out_size - tindex)[j+i];
				tmp_i[tindex] = cInl->getPrevious(out_size - tindex)[j+i];
				tmp_u[tindex] = cVnl->getPrevious(out_size - tindex)[j+i];
      }
      elem_vec[k]->getElemData()->setRealX(j, tmp_i);
      elem_vec[k]->getElemData()->setRealI(j, tmp_i);
      elem_vec[k]->getElemData()->setRealU(j, tmp_u);
    }
    i += current_ns;
  }

  if (savenode) {
    // For each terminal, assign voltage vector
    Terminal* term = NULL;
    cir->setFirstTerminal();
    while((term = cir->nextTerminal())) {
      // Get MNAM index
      if (term->getRC()) {
				// Remember that MNAM indices begin at 1
				i = term->getRC() - 1;
				for (int tindex=0; tindex < out_size; tindex++)
					tmp_u[tindex] = cU->getPrevious(out_size - tindex)[i];
      }
      else
				// This is a reference terminal
			mtl::set_value(tmp_u, zero);

      // Set terminal vector
      term->getTermData()->setRealV(tmp_u);
    }
  }
}
