#include "WTDBase.h"

extern "C"
{
	#include "../inout/ftvec.h"
	#include "../inout/report.h"
}

WTDBase::WTDBase(ElementVector& elem_vec, const int& nstates,
const int& nwindows)
: elem_vec(elem_vec), nstates(nstates),
nwindows(nwindows), nw(0), nsamples(0), lastsample(0)
{
  timev = new DenseDoubleVector[nwindows];
  xw = new DenseDoubleVector[nwindows];
  vpw = new DenseDoubleVector[nwindows];
  ipw = new DenseDoubleVector[nwindows];
  uw = new DenseDoubleVector[1];
}

WTDBase::~WTDBase()
{
  delete [] uw;
  delete [] ipw;
  delete [] vpw;
  delete [] xw;
  delete [] timev;
}

void WTDBase::saveWindow(WTLinear *wtl, ExtDoubleVector& x,
DenseDoubleVector& vp, DenseDoubleVector ip)
{
  // get time vector for this window
  timev[nw] = DenseDoubleVector(nsamples);
  wtl->getTimeVector(timev[nw]);
  // Copy output vectors
  xw[nw] = DenseDoubleVector(nsamples * nstates);
  copy(x, xw[nw]);
  vpw[nw] = DenseDoubleVector(nsamples * nstates);
  copy(vp, vpw[nw]);
  ipw[nw] = DenseDoubleVector(nsamples * nstates);
  copy(ip, ipw[nw]);
  // Calculate the response of the linear part
  uw[0] = DenseDoubleVector(nsamples * wtl->getNLinVars());
  wtl->getU(ip, uw[0]);
}


void WTDBase::getXIC(DenseDoubleVector& x_ic)
{
  // #########################################################
  // Warning: this only works with current wavelet base
  // Otherwiae we must store the actual samples or do the
  // transformation before returning the samples.
  // #########################################################

  // Return last sample from previous window (or 0 if nw == 0)
  if (nw) {
    int prev_nsamples = timev[nw-1].size();
    for (int i=0; i < nstates; i++)
      x_ic[i] = xw[nw-1][i * prev_nsamples + lastsample];
  }
  else {
    for (int i=0; i < nstates; i++)
      x_ic[i] = zero;
  }
}

void WTDBase::getUIC(DenseDoubleVector& u_ic)
{
  int nlinvars = u_ic.size();
  // Return last sample from previous window (or 0 if nw == 0)
  if (nw) {
    int prev_nsamples = timev[nw-1].size();
    for (int i=0; i < nlinvars; i++)
      u_ic[i] = uw[0][i * prev_nsamples + lastsample];
  }
  else {
    for (int i=0; i < nlinvars; i++)
      u_ic[i] = zero;
  }
}

void WTDBase::writeOutput()
{
  report(MESSAGE, "--- Writing output vectors ...");

  // This routine should save all the windows that have been set.
  assert(nw <= nwindows);

  // Find out how many time samples are there
  int allsamples(0);
  for (int i(0); i < nw+1; i++)
    allsamples += timev[i].size();

  // Allocate and copy global time vector
  allocTimeV_P(allsamples);
  int tindex(0);
  for (int i(0); i < nw+1; i++)
	for (unsigned j(0); j < timev[i].size(); j++) {
		TimeV_P[tindex] = timev[i][j];
		tindex++;
	}

  // Now fill currents and port voltages of time domain devices.
  int n_elem = elem_vec.size();   // Number of elements
  int i=0;
  DenseDoubleVector dx(allsamples);
  DenseDoubleVector dv(allsamples);
  DenseDoubleVector di(allsamples);
  for (int k=0; k < n_elem; k++) {
    int current_ns = elem_vec[k]->getNumberOfStates();
    for(int j=0; j< current_ns; j++) {
      tindex = 0;
      for (int m(0); m < nw+1; m++) {
				int cnsamples = timev[m].size();
				for (int l(0); l < cnsamples; l++) {
					dx[tindex] = xw[m][(i+j) * cnsamples + l];
					dv[tindex] = vpw[m][(i+j) * cnsamples + l];
					di[tindex] = ipw[m][(i+j) * cnsamples + l];
					tindex++;
				}
      }
      elem_vec[k]->getElemData()->setRealX(j, dx);
      elem_vec[k]->getElemData()->setRealU(j, dv);
      elem_vec[k]->getElemData()->setRealI(j, di);
    }
    i += current_ns;
  }
}

