// This class handles the linear system in the Wavelet transient.
//
// Author:
//          Carlos E. Christoffersen
//

#ifndef WTLinear_h
#define WTLinear_h 1

#include "WaveletBase.h"
#include "TimeMNAM.h"
#include "../network/Circuit.h"
extern "C" {
	#undef _S
	#undef _C
	#include <dsp_defs.h>
	#include <util.h>
}

// This structure contains all the variables needed for a fortran-style
// sparse matrix.
// This struct is intended to be used inside WTLinear only.
struct SPStorage {

  SPStorage();
  ~SPStorage();
  void updateSize(const int& ncols, const int& nnz);
  inline const int& getNCols() const
  {
    return ncols;
  }
  inline const int& getNNZ() const
  {
    return nnz;
  }

  // Pointers are public to ease the interface with fortran routines.
  // Use with care!
  int *col_pointer;
  int *row_index;
  double *elem_val;
  double *rhs;
  double *x;

	private:
  int ncols;
  int nnz;
};

// Main class definition follows

class WTLinear
{
	public:

  WTLinear();

  ~WTLinear();

  // Setup the class
  void setup(WaveletBase* wb, TimeMNAM* mnam, const IntDenseColMatrix& T,
	DenseDoubleVector& icv, int& permc_spec);

  // Write the expanded MNAM and vectors in HB format
  void writeHB();

  // Set the time window start (update source vector)
  inline void setTWStart(const double& ts, DenseDoubleVector& icv)
  {
    update_svw(ts, icv);
    update_ssv();
  }

  // Scale the compressed source vector by a factor (for source stepping)
  void scaleSsv(const double& factor);

  // get vL given iNL
  void getVL(const DenseDoubleVector& iNL, DenseDoubleVector& vL);

  // add the `linear' Jacobian to Jac
  void getJvL(DoubleDenseColMatrix& JiNL, DoubleDenseExtColMatrix& Jac);

  void addJvLBlock(DoubleDenseColMatrix& JIp, DoubleDenseExtColMatrix& Jac,
	const int& blocksize, const int& offset);

  // Returns the number of circuit variables in the linear part
  inline int getNLinVars()
  {
    return mnam->getDim();
  }

  // copy the time vector used
  void getTimeVector(DenseDoubleVector& timev);

  // get a row of the `linear' Jacobian
  // void getJvLrow(const SparseMatrix& JiNL, DenseVector& JvLrow);

  // returns the vector of unknowns given the nonlinear current vector
  void getU(const DenseDoubleVector& iNL, DenseDoubleVector& uw);

	private:

  // Update the source vector
  void update_svw(const double& ts, DenseDoubleVector& icv);
  void update_ssv();

  bool ready;
  WaveletBase* wb;
  TimeMNAM* mnam;
  IntDenseColMatrix T;
  DenseDoubleVector time_vec;

  DoubleSparseColMatrix eWj;
  int n_states;
  int eblksize;
  int nsamples;

  DenseDoubleVector svw;
  DoubleDenseColMatrix Msv;
  DenseDoubleVector ssv;
  DenseDoubleVector backup_ssv;

  // spstor is to hold the big matrix.
  SPStorage spstor;

  // SuperLU variables
  SuperMatrix A, L, U, B, X;
  int *perm_r;
  int *perm_c;
  int *etree;
  double *r, *c;
  void *work;
  int lwork;

};

#endif

