// This class provides all the methods related with the particular
// wavelet base used by the SVWav analysis.
//
// Author:
//         Carlos E. Christoffersen
//

#ifndef WaveletBase_h
#define WaveletBase_h 1

#include "../freeda_mtl.h"
// #include "../network/standard.h"
extern const double zero;
extern const double one;

// This class handle all conversions between coefficient and physical
// spaces.
class WaveletBase
{
	public:

  // J: initial subspace level
  // tw: time window width
  WaveletBase(const int& L, const int& J);

  ~WaveletBase();

  // Set up transformation matrices
  void setup(const double& tw);

  // Modify time window width
  void setTimeWindow(const double& tw);

  // Get the size of the transformation matrix
  inline int getSize()
  {
    return size;
  }

  // Get the index of the last sample in the interval.
  inline int getLastIndex()
  {
    return L+2;
  }

  // Returns a vector with all the time points in use for each variable
  void getTimePointsVector(DenseDoubleVector& result);

  // returns WJ
  void getWJ(DoubleSparseColMatrix& wj);

  // Returns W'J
  void getWpJ(DoubleSparseColMatrix& wpj);

  // returns WJr for Jacobian calculations
  //
  // WJr = WJ(1:n, 1:n) - 1/WJ(0,0) * WJ(1:n, 0) * WJ(0, 1:n)
  //
  void getWJr(DoubleSparseColMatrix& ewj);

  // Returns W'Jr for Jacobian calculations
  //
  // WpJr = WpJ(1:n, 1:n) - 1/WJ(0,0) * WpJ(1:n, 0) * WJ(0, 1:n)
  //
  void getWpJr(DoubleSparseColMatrix& ewpj);

  // Convert to time domain from reduced coefficient vector
  // and initial condition
  void toTime(const double& x0,
	const ExtDoubleVector& xhat_r,
	DenseDoubleVector& x_r);

  // Derivate from reduced coefficient vector
  // and initial condition
  void toTimeDeriv(const double& x0,
	const ExtDoubleVector& xhat_r,
	DenseDoubleVector& xd_r);

  // Get a guess coefficient vector given the initial condition
  void getGuess(const double& x0, ExtDoubleVector& x_r);

	private:

  // Auxiliary functions
  inline double xp(const double& x)
	{
		return (x > zero) ? x : zero;
	}

  double fik(const double &x, const int &k);

  double psijk(const double &x, const int& j, const int &k, double nj);

  double dfik(const double &x, const int &k);

  double dpsijk(const double &x, const int& j, const int &k, double nj);

  // Scaling functions
  double n1(const double &x);

  double n2(const double &x);

  double phib(const double &x);

  double phi(const double &x);

  // Wavelet functions
  double psib0(const double &x);

  double psib1(const double &x);

  double psi(const double &x);

  // Derivatives of scaling and wavelet functions

  // Scaling functions (Derivatives)
  double dn1(const double &x);

  double dn2(const double &x);

  double dphib(const double &x);

  double dphi(const double &x);

  // Wavelet functions (Derivatives)
  double dpsib0(const double &x);

  double dpsib1(const double &x);

  double dpsi(const double &x);

  // L in Cai's method
  int L;
  // Double version of L
  double dbL;
  // J in Cai's method
  int J;
  // Time window size
  double tw;

  // Size of matrices and vectors
  int size;
  // Expansion size
  int esize;
  // Transformation and derivative (and derivative scaled) matrices
  DoubleSparseColMatrix Wj, Wpj, Wpsj, WJr, WpsJr;
  // Collocation points vector
  DenseDoubleVector cPoints;
  // Scaled collocation points vector
  DenseDoubleVector scPoints;
  // Wj(1:n-1, 0) and Wpj(1:n-1, 0)
  DenseDoubleVector c0r_Wj, c0r_Wpj;
  // Coefficients for a constant function = 1.
  DenseDoubleVector unit_xh;
  // Flag to tell if the matrices are ready
  bool ready;
};

#endif

