#include "../../../../network/ElementManager.h"
#include "../../../../network/AdolcElement.h"
#include "CmosInv.h"

// Define the number of parameters
const unsigned CmosInv :: n_par = 12;

// Element Information
ItemInfo CmosInv::einfo =
{
	"cmosinv",
	"CMOS Inverter",
	"Mazen M Kharbutli",
	DEFAULT_ADDRESS"category:cmos,digital"
};

// Parameter Information
ParmInfo CmosInv::pinfo[] =
{
	{"vtn","NMOS Threshold Voltage (V)", TR_DOUBLE, false},
	{"vtp","PMOS Threshold Voltage (V)", TR_DOUBLE, false},
	{"un","Effective Mobility of Electrons in NMOS ((cm^2)/(V-sec))", TR_DOUBLE, false},
	{"up","Effective Mobility of Holes in PMOS ((cm^2)/(V-sec))", TR_DOUBLE, false},
	{"en","Permittivity of the Gate Insultor in NMOS (F/cm)", TR_DOUBLE, false},
	{"ep","Permittivity of the Gate Insultor in PMOS (F/cm)", TR_DOUBLE, false},
	{"tox","Thickness of the Gate Insulator (cm)", TR_DOUBLE, false},
	{"wn","Channel Width of NMOS (cm)", TR_DOUBLE, false},
	{"ln","Channel Length of NMOS (cm)", TR_DOUBLE, false},
	{"wp","Channel Width of PMOS (cm)", TR_DOUBLE, false},
	{"lp","Channel Length of PMOS (cm)", TR_DOUBLE, false},
	{"td","Response Delay Time (sec)", TR_DOUBLE, false},
};

// Constructor
CmosInv::CmosInv(const string& iname)
:AdolcElement(&einfo, pinfo, n_par, iname)
{
	// Set default parameter values
	paramvalue[0] = &(vtn = 1);
	paramvalue[1] = &(vtp = -1);
	paramvalue[2] = &(un = 500);
	paramvalue[3] = &(up = 200);
	paramvalue[4] = &(en = 34.515e-14);
	paramvalue[5] = &(ep = 34.515e-14);
	paramvalue[6] = &(tox = 2e-6);
	paramvalue[7] = &(ln = 2.0e-6);
	paramvalue[8] = &(wn = 50e-6);
	paramvalue[9] = &(lp = 2.0e-6);
	paramvalue[10] = &(wp = 50e-6);
	paramvalue[11] = &(td = 0);

	// Set Number of Terminals
	setNumTerms(4);

	// Set Flags
	setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);

	// Set Number of States
	setNumberOfStates(3);
}

// Initialization Function
void CmosInv::init() throw(string&)
{
	// Find Beta for the NMOS and PMOS
	betan = (un * en * wn) / (tox * ln);
	betap = (up * ep * wp) / (tox * lp);

	//create tape
	DenseIntVector var(3);
	var[0] = 0;
	var[1] = 1;
	var[2] = 2;
	DenseIntVector dvar(3);
	dvar[0] = 0;
	dvar[1] = 1;
	dvar[2] = 2;
	DenseIntVector none(3,0);
	DenseIntVector tvar(3);
	tvar[0] = 0;
	tvar[1] = 1;
	tvar[2] = 2;
	DenseDoubleVector delay(3);
	delay[0] = 0;
	delay[1] = td;
	delay[2] = 0;
	createTape(var, dvar, none, tvar, delay);
}

// Evaluate Function
void CmosInv::eval(adoublev& x, adoublev& vp, adoublev& ip)
{
	//x[0]  : Vdd    x[3] : d(Vdd)/dt
	//x[1]  : Vin    x[4] : d(Vin)/dt
	//x[2]  : Vout   x[5] : d(Vout)/dt
	//x[10] : Vout(t-td)
	//vp[0] : Vdd   ip[0] : Isdp
	//vp[1] : Vin   ip[1] : Ig
	//vp[2] : Vout  ip[2] : Iout

	adouble Idsn, Idsp, Vout, VoutTime;
	adouble IdsnB, IdspB, VoutB, tempB;
	adouble IdsnC, IdspC, VoutC, tempC;
	adouble IdsnD, IdspD, VoutD, tempD;

	// Region A: 0 <= Vin <= Vtn
	Vout = x[0];
	Idsn = 0;
	Idsp = 0;

	// Region B: Vtn <= Vin <= Vdd/2
	IdsnB = (betan/2) * (x[10] - vtn) * (x[10] - vtn);
	IdspB = -betap * (((x[10] - x[0] - vtp) * (x[2] - x[0]))
	- ((x[2] - x[0]) * (x[2] - x[0]) / 2));
	tempB = (x[10] - vtp)*(x[10] - vtp);
	tempB = tempB - (2 * x[0] * (x[10] - (x[0]/2) - vtp));
	tempB = tempB - ((betan/betap) * (x[10] - vtn) * (x[10] - vtn));
	VoutB = (x[10] - vtp) + sqrt(tempB);

	condassign(Vout,(x[10]-vtn),VoutB,Vout);
	condassign(Idsn,(x[10]-vtn),IdsnB,Idsn);
	condassign(Idsp,(x[10]-vtn),IdspB,Idsp);

	// Region C: Vin = Vdd/2
	IdsnC = betan * (x[10] - vtn) * (x[10] - vtn) / 2;
	IdspC = -IdsnC;
	VoutC = x[2];

	//condassign(Vout,(x[1]-(x[0]/2)),VoutC,Vout);
	//condassign(Idsn,(x[1]-(x[0]/2)),IdsnC,Idsn);
	//condassign(Idsp,(x[1]-(x[0]/2)),IdspC,Idsp);

	// Region D: Vdd/2 <= Vin <= Vdd + Vtp
	IdsnD = betan * (((x[10] - vtn) * x[2]) - (x[2] * x[2] / 2));
	IdspD = - (betap/2) * (x[10] - x[0] - vtp) * (x[10] - x[0] - vtp);
	tempD = (x[10] - vtn) * (x[10] - vtn);
	tempD = tempD - ((betap/betan) * (x[10] - x[0] - vtp) * (x[10] - x[0] - vtp));
	VoutD = (x[10] - vtn) - sqrt(tempD);

	condassign(Vout,(x[10]-(x[0]/2)),VoutD,Vout);
	condassign(Idsn,(x[10]-(x[0]/2)),IdsnD,Idsn);
	condassign(Idsp,(x[10]-(x[0]/2)),IdspD,Idsp);

	// Region E: Vin >= Vdd + Vtp
	condassign(Vout,(x[10]-x[0]-vtp),0,Vout);
	condassign(Idsn,(x[10]-x[0]-vtp),0,Idsn);
	condassign(Idsp,(x[10]-x[0]-vtp),0,Idsp);

	vp[0] = x[0];
	vp[1] = x[1];
	vp[2] = Vout;

	ip[0] = -Idsp;
	ip[1] = 0;
	ip[2] = Idsn + Idsp;
}

