#include "../../../../network/ElementManager.h"
#include "../../../../network/AdolcElement.h"
#include "CmosNand.h"

// Define the number of parameters
const unsigned CmosNand :: n_par = 12;

// Element Information
ItemInfo CmosNand::einfo =
{
	"cmosnand",
	"CMOS NAND Gate",
	"Mazen M Kharbutli",
	DEFAULT_ADDRESS"category:cmos,digital",
	"2003_05_15"
};

// Parameter Information
ParmInfo CmosNand::pinfo[] =
{
	{"vtn","NMOS Threshold Voltage (V)", TR_DOUBLE, false},
	{"vtp","PMOS Threshold Voltage (V)", TR_DOUBLE, false},
	{"un","Effective Mobility of Electrons in NMOS ((cm^2)/(V-sec))", TR_DOUBLE, false},
	{"up","Effective Mobility of Holes in PMOS ((cm^2)/(V-sec))", TR_DOUBLE, false},
	{"en","Permittivity of the Gate Insultor in NMOS (F/cm)", TR_DOUBLE, false},
	{"ep","Permittivity of the Gate Insultor in PMOS (F/cm)", TR_DOUBLE, false},
	{"tox","Thickness of the Gate Insulator (cm)", TR_DOUBLE, false},
	{"wn","Channel Width of NMOS (cm)", TR_DOUBLE, false},
	{"ln","Channel Length of NMOS (cm)", TR_DOUBLE, false},
	{"wp","Channel Width of PMOS (cm)", TR_DOUBLE, false},
	{"lp","Channel Length of PMOS (cm)", TR_DOUBLE, false},
	{"td","Response Delay Time (sec)", TR_DOUBLE, false},
};

// Constructor
CmosNand::CmosNand(const string& iname)
:AdolcElement(&einfo, pinfo, n_par, iname)
{
	// Set default parameter values
	paramvalue[0] = &(vtn = 1);
	paramvalue[1] = &(vtp = -1);
	paramvalue[2] = &(un = 500);
	paramvalue[3] = &(up = 200);
	paramvalue[4] = &(en = 34.515e-14);
	paramvalue[5] = &(ep = 34.515e-14);
	paramvalue[6] = &(tox = 2e-6);
	paramvalue[7] = &(ln = 2.0e-6);
	paramvalue[8] = &(wn = 50e-6);
	paramvalue[9] = &(lp = 2.0e-6);
	paramvalue[10] = &(wp = 50e-6);
	paramvalue[11] = &(td = 0);

	// Set Number of Terminals
	setNumTerms(5);

	// Set Flags
	setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);

	// Set Number of States
	setNumberOfStates(4);
}

// Initialization Function
void CmosNand::init() throw(string&)
{
	// Find Beta for the NMOS and PMOS
	betan = (un * en * wn) / (tox * ln);
	betap = (up * ep * wp) / (tox * lp);

	//create tape
	DenseIntVector var(4);
	var[0] = 0;
	var[1] = 1;
	var[2] = 2;
	var[3] = 3;
	DenseIntVector dvar(4);
	dvar[0] = 0;
	dvar[1] = 1;
	dvar[2] = 2;
	dvar[3] = 3;
	DenseIntVector none(4,0);
	DenseIntVector tvar(4);
	tvar[0] = 0;
	tvar[1] = 1;
	tvar[2] = 2;
	tvar[3] = 3;
	DenseDoubleVector delay(4);
	delay[0] = 0;
	delay[1] = td;
	delay[2] = td;
	delay[3] = 0;
	createTape(var, dvar, none, tvar, delay);
}

// Evaluate Function
void CmosNand::eval(adoublev& x, adoublev& vp, adoublev& ip)
{
	// x[0]  : Vdd    x[4] : d(Vdd)/dt
	// x[1]  : Vin1   x[5] : d(Vin1)/dt
	// x[2]  : Vin2   x[6] : d(Vin2)/dt
	// x[3]  : Vout   x[7] : d(Vout)/dt
	// vp[0] : Vdd    ip[0] : Isdp
	// vp[1] : Vin1   ip[1] : Ig1
	// vp[2] : Vin2   ip[2] : Ig2
	// vp[3] : Vout   ip[3] : Iout

	adouble Idsn, Idsp, Vout, VoutTime;
	adouble Idsn1, Idsp1, Vout1;
	adouble IdsnA1, IdspA1, VoutA1;
	adouble IdsnB1, IdspB1, VoutB1, tempB1;
	adouble IdsnC1, IdspC1, VoutC1, tempC1;
	adouble IdsnD1, IdspD1, VoutD1, tempD1;
	adouble IdsnE1, IdspE1, VoutE1;
	adouble Idsn2, Idsp2, Vout2;
	adouble IdsnA2, IdspA2, VoutA2;
	adouble IdsnB2, IdspB2, VoutB2, tempB2;
	adouble IdsnC2, IdspC2, VoutC2, tempC2;
	adouble IdsnD2, IdspD2, VoutD2, tempD2;
	adouble IdsnE2, IdspE2, VoutE2;

	// Region A1: 0 <= Vin1 <= Vtn
	Vout1 = x[0];
	Idsn1 = 0;
	Idsp1 = 0;

	// Region B1: Vtn <= Vin1 <= Vdd/2
	IdsnB1 = (betan/2) * (x[13] - vtn) * (x[13] - vtn);
	IdspB1 = -betap * (((x[13] - x[0] - vtp) * (x[3] - x[0]))
	- ((x[3] - x[0]) * (x[3] - x[0]) / 2));
	tempB1 = (x[13] - vtp)*(x[13] - vtp);
	tempB1 = tempB1 - (2 * x[0] * (x[13] - (x[0]/2) - vtp));
	tempB1 = tempB1 - ((betan/betap) * (x[13] - vtn) * (x[13] - vtn));
	VoutB1 = (x[13] - vtp) + sqrt(tempB1);

	condassign(Vout1,(x[13]-vtn),VoutB1,Vout1);
	condassign(Idsn1,(x[13]-vtn),IdsnB1,Idsn1);
	condassign(Idsp1,(x[13]-vtn),IdspB1,Idsp1);

	// Region C1: Vin1 = Vdd/2
	IdsnC1 = betan * (x[13] - vtn) * (x[13] - vtn) / 2;
	IdspC1 = -IdsnC1;
	VoutC1 = x[3];

	// Region D1: Vdd/2 <= Vin1 <= Vdd + Vtp
	IdsnD1 = betan * (((x[13] - vtn) * x[3]) - (x[3] * x[3] / 2));
	IdspD1 = - (betap/2) * (x[13] - x[0] - vtp) * (x[13] - x[0] - vtp);
	tempD1 = (x[13] - vtn) * (x[13] - vtn);
	tempD1 = tempD1 - ((betap/betan) * (x[13] - x[0] - vtp) * (x[13] - x[0] - vtp));
	VoutD1 = (x[13] - vtn) - sqrt(tempD1);

	condassign(Vout1,(x[13]-(x[0]/2)),VoutD1,Vout1);
	condassign(Idsn1,(x[13]-(x[0]/2)),IdsnD1,Idsn1);
	condassign(Idsp1,(x[13]-(x[0]/2)),IdspD1,Idsp1);

	// Region E1: Vin1 >= Vdd + Vtp
	condassign(Vout1,(x[13]-x[0]-vtp),0,Vout1);
	condassign(Idsn1,(x[13]-x[0]-vtp),0,Idsn1);
	condassign(Idsp1,(x[13]-x[0]-vtp),0,Idsp1);

	// Region A2: 0 <= Vin2 <= Vtn
	Vout2 = x[0];
	Idsn2 = 0;
	Idsp2 = 0;

	// Region B2: Vtn <= Vin2 <= Vdd/2
	IdsnB2 = (betan/2) * (x[14] - vtn) * (x[14] - vtn);
	IdspB2 = -betap * (((x[14] - x[0] - vtp) * (x[3] - x[0]))
	- ((x[3] - x[0]) * (x[3] - x[0]) / 2));
	tempB2 = (x[14] - vtp)*(x[14] - vtp);
	tempB2 = tempB2 - (2 * x[0] * (x[14] - (x[0]/2) - vtp));
	tempB2 = tempB2 - ((betan/betap) * (x[14] - vtn) * (x[14] - vtn));
	VoutB2 = (x[14] - vtp) + sqrt(tempB2);

	condassign(Vout2,(x[14]-vtn),VoutB2,Vout2);
	condassign(Idsn2,(x[14]-vtn),IdsnB2,Idsn2);
	condassign(Idsp2,(x[14]-vtn),IdspB2,Idsp2);

	// Region C2: Vin2 = Vdd/2
	IdsnC2 = betan * (x[14] - vtn) * (x[14] - vtn) / 2;
	IdspC2 = -IdsnC2;
	VoutC2 = x[3];

	// Region D2: Vdd/2 <= Vin2 <= Vdd + Vtp
	IdsnD2 = betan * (((x[14] - vtn) * x[3]) - (x[3] * x[3] / 2));
	IdspD2 = - (betap/2) * (x[14] - x[0] - vtp) * (x[14] - x[0] - vtp);
	tempD2 = (x[14] - vtn) * (x[14] - vtn);
	tempD2 = tempD2 - ((betap/betan) * (x[14] - x[0] - vtp) * (x[14] - x[0] - vtp));
	VoutD2 = (x[14] - vtn) - sqrt(tempD2);

	condassign(Vout2,(x[14]-(x[0]/2)),VoutD2,Vout2);
	condassign(Idsn2,(x[14]-(x[0]/2)),IdsnD2,Idsn2);
	condassign(Idsp2,(x[14]-(x[0]/2)),IdspD2,Idsp2);

	// Region E2: Vin2 >= Vdd + Vtp
	condassign(Vout2,(x[14]-x[0]-vtp),0,Vout2);
	condassign(Idsn2,(x[14]-x[0]-vtp),0,Idsn2);
	condassign(Idsp2,(x[14]-x[0]-vtp),0,Idsp2);

	// Assign output voltages
	vp[0] = x[0];
	vp[1] = x[1];
	vp[2] = x[2];
	condassign(Vout,(x[14]-x[13]),Vout1,Vout2);
	vp[3] = Vout;

	// Assign output current
	condassign(Idsp,(x[14]-x[13]),Idsp1,Idsp2);
	condassign(Idsn,(x[14]-x[13]),Idsn1,Idsn2);
	ip[0] = -Idsp1 - Idsp2;
	ip[1] = 0;
	ip[2] = 0;
	ip[3] = Idsn + Idsp1 + Idsp2;
}

