// Thermal CMOS NOR Gate
// By Travis Lentz and Tony Mulder

#include "../../../../network/ElementManager.h"
#include "../../../../network/AdolcElement.h"
#include "CmosNorT.h"

// Define the number of parameters
const unsigned CmosNorT :: n_par = 23;

// Element Information
ItemInfo CmosNorT::einfo = {
   "cmosnort",
   "CMOS NOR Gate, Electro-thermal",
   "Travis Lentz, Tony Mulder",
   DEFAULT_ADDRESS"category:cmos,electrothermal,digital",
   "2005_04_19"
};

// Parameter Information
ParmInfo CmosNorT::pinfo[] = {
   {"vtn","NMOS Threshold Voltage (V)", TR_DOUBLE, false},
   {"vtp","PMOS Threshold Voltage (V)", TR_DOUBLE, false},
   {"un","Effective Mobility of Electrons in NMOS ((cm^2)/(V-sec))", TR_DOUBLE, 
false},
   {"up","Effective Mobility of Holes in PMOS ((cm^2)/(V-sec))", TR_DOUBLE, false},    
   {"en","Permittivity of the Gate Insultor in NMOS (F/cm)", TR_DOUBLE, false},
   {"ep","Permittivity of the Gate Insultor in PMOS (F/cm)", TR_DOUBLE, false},
   {"tox","Thickness of the Gate Insulator (cm)", TR_DOUBLE, false},
   {"wn","Channel Width of NMOS (cm)", TR_DOUBLE, false},
   {"ln","Channel Length of NMOS (cm)", TR_DOUBLE, false}, 
   {"wp","Channel Width of PMOS (cm)", TR_DOUBLE, false},
   {"lp","Channel Length of PMOS (cm)", TR_DOUBLE, false},
   {"td","Response Delay Time (sec)", TR_DOUBLE, false},
   {"thermal", "Thermal element flag", TR_BOOLEAN, false},
   {"ts", "Kirchhoff transformation temperature (K)", TR_DOUBLE, false}, 
   {"tnom", "Ambient temperature (K)", TR_DOUBLE, false},
   {"zt", "Thermal impedence summation (K/W)", TR_DOUBLE, false},
   {"c1", "P channel GS capacitance (F)", TR_DOUBLE, false},
   {"c2", "N channel GS capacitance (F)", TR_DOUBLE, false},
   {"c3", "output GS (Miller) capacitance (F)", TR_DOUBLE, false},
   {"c4", "parasitic diode capacitance from out to Vdd (F)", TR_DOUBLE, false},
   {"c5", "parasitic diode capacitance from out to Gnd(F)", TR_DOUBLE, false},
   {"freq", "operating frequency (Hz)", TR_DOUBLE, false},
   {"lk", "leakage current (A)", TR_DOUBLE, false}
};

// Constructor
CmosNorT::CmosNorT(const string& iname)
   :AdolcElement(&einfo, pinfo, n_par, iname)
{
   // Set default parameter values
   paramvalue[0] = &(vtn = 1);
   paramvalue[1] = &(vtp = -1);
   paramvalue[2] = &(un = 500);
   paramvalue[3] = &(up = 200);
   paramvalue[4] = &(en = 34.515e-14);
   paramvalue[5] = &(ep = 34.515e-14);   
   paramvalue[6] = &(tox = 2e-6);
   paramvalue[7] = &(ln = 2.0e-6);
   paramvalue[8] = &(wn = 50e-6);
   paramvalue[9] = &(lp = 2.0e-6);
   paramvalue[10] = &(wp = 100e-6);
   paramvalue[11] = &(td = 0);
   paramvalue[12] = &(thermal = false);
   paramvalue[13] = &(ts = 300.);
   paramvalue[14] = &(tnom = 300);
   paramvalue[15] = &(zt = 310);
   paramvalue[16] = &(c1 = 1.0e-12);
   paramvalue[17] = &(c2 = 1.0e-12);
   paramvalue[18] = &(c3 = 1.0e-12);
   paramvalue[19] = &(c4 = 1.0e-12);
   paramvalue[20] = &(c5 = 1.0e-12);
   paramvalue[21] = &(freq = 1.0e6);
   paramvalue[22] = &(lk = 8e-6);

   // Set Number of Terminals
   setNumTerms(7);

   // Set Flags
   setFlags(NONLINEAR | MULTI_REF | TR_TIME_DOMAIN);

   // Set Number of States
   setNumberOfStates(5);

}

// Initialization Function
void CmosNorT::init() throw(string&)
{
   // Find Beta for the NMOS and PMOS
   betan = (un * en * wn) / (tox * ln);
   betap = (up * ep * wp) / (tox * lp);
   power = ( freq * (c1+c2+c3+c4+c5));

   //create tape
   DenseIntVector var(5);
   var[0] = 0;
   var[1] = 1;
   var[2] = 2;
   var[3] = 3;
   var[4] = 4;
   DenseIntVector dvar(5);
   dvar[0] = 0;
   dvar[1] = 1;
   dvar[2] = 2;
   dvar[3] = 3;
   dvar[4] = 4;
   DenseIntVector none(5,0);
   DenseIntVector tvar(5);
   tvar[0] = 0;
   tvar[1] = 1;
   tvar[2] = 2;
   tvar[3] = 3;
   tvar[4] = 4;
   DenseDoubleVector delay(5);
   delay[0] = 0;
   delay[1] = td;
   delay[2] = td;
   delay[3] = 0;
   delay[4] = 0;
   createTape(var, dvar, none, tvar, delay);
}

void CmosNorT::getLocalRefIdx(UnsignedVector& local_ref_vec,
			      TerminalVector& term_list)
{
  // Make sure the vectors are empty
  term_list.erase(term_list.begin(), term_list.end());
  local_ref_vec.erase(local_ref_vec.begin(), local_ref_vec.end());

  // Insert vector elements
  term_list.push_back(getTerminal(0));
  term_list.push_back(getTerminal(1));
  term_list.push_back(getTerminal(2));
  term_list.push_back(getTerminal(3));
  term_list.push_back(getTerminal(4)); // Local reference terminal
  term_list.push_back(getTerminal(5));
  term_list.push_back(getTerminal(6)); // Local reference terminal
  
  local_ref_vec.push_back(4); // Local reference index
  local_ref_vec.push_back(6); // Local reference index
}

// Evaluate Function
void CmosNorT::eval(adoublev& x, adoublev& vp, adoublev& ip)
{
    // x[0]  : Vdd    x[5] : d(Vdd)/dt
    // x[1]  : Vin1   x[6] : d(Vin1)/dt
    // x[2]  : Vin2   x[7] : d(Vin2)/dt
    // x[3]  : Vout   x[8] : d(Vout)/dt
    // x[4]  : temp
    // vp[0] : Vdd    ip[0] : Isdp
    // vp[1] : Vin1   ip[1] : Ig1
    // vp[2] : Vin2   ip[2] : Ig2
    // vp[3] : Vout   ip[3] : Iout
    // vp[4] : temp   ip[4] : pout
    
    adouble Idsn, Idsp, Vout, VoutTime, tj, pwr;
    adouble Idsn1, Idsp1, Vout1;
    adouble IdsnA1, IdspA1, VoutA1;
    adouble IdsnB1, IdspB1, VoutB1, tempB1;
    adouble IdsnC1, IdspC1, VoutC1, tempC1;
    adouble IdsnD1, IdspD1, VoutD1, tempD1;    
    adouble IdsnE1, IdspE1, VoutE1;
    adouble Idsn2, Idsp2, Vout2;
    adouble IdsnA2, IdspA2, VoutA2;
    adouble IdsnB2, IdspB2, VoutB2, tempB2;
    adouble IdsnC2, IdspC2, VoutC2, tempC2;
    adouble IdsnD2, IdspD2, VoutD2, tempD2;    
    adouble IdsnE2, IdspE2, VoutE2;
    
     // Region A1: 0 <= Vin1 <= Vtn
    Vout1 = x[0];
    Idsn1 = lk;
    Idsp1 = -lk;
   
    // Region B1: Vtn <= Vin1 <= Vdd/2 
    IdsnB1 = ((betan/2) * (x[16] - vtn) * (x[16] - vtn))+(power*x[0])+lk;
    IdspB1 = (-betap * (((x[16] - x[0] - vtp) * (x[3] - x[0])) 
      - ((x[3] - x[0]) * (x[3] - x[0]) / 2)))-(power*x[0])-lk;
    tempB1 = (x[16] - vtp)*(x[16] - vtp);
    tempB1 = tempB1 - (2 * x[0] * (x[16] - (x[0]/2) - vtp));
    tempB1 = tempB1 - ((betan/betap) * (x[16] - vtn) * (x[16] - vtn));
    VoutB1 = (x[16] - vtp) + sqrt(tempB1);
     
    condassign(Vout1,(x[16]-vtn),VoutB1,Vout1);
    condassign(Idsn1,(x[16]-vtn),IdsnB1,Idsn1);
    condassign(Idsp1,(x[16]-vtn),IdspB1,Idsp1);
   
    // Region C1: Vin1 = Vdd/2
    IdsnC1 = (betan * (x[16] - vtn) * (x[16] - vtn) / 2)+(power*x[0])+lk;
    IdspC1 = -IdsnC1;
    VoutC1 = x[3];
    
    // Region D1: Vdd/2 <= Vin1 <= Vdd + Vtp 
    IdsnD1 = (betan * (((x[16] - vtn) * x[3]) - (x[3] * x[3] / 2))) +
     (power*x[0])+lk;
    IdspD1 = (- (betap/2) * (x[16] - x[0] - vtp) * (x[16] - x[0] - vtp))
         - (power*x[0])-lk;
    tempD1 = (x[16] - vtn) * (x[16] - vtn);
    tempD1 = tempD1 - ((betap/betan) * (x[16] - x[0] - vtp) * (x[16] - x[0] - vtp));
    VoutD1 = (x[16] - vtn) - sqrt(tempD1);

    condassign(Vout1,(x[16]-(x[0]/2)),VoutD1,Vout1);
    condassign(Idsn1,(x[16]-(x[0]/2)),IdsnD1,Idsn1);
    condassign(Idsp1,(x[16]-(x[0]/2)),IdspD1,Idsp1);
    
    // Region E1: Vin1 >= Vdd + Vtp
    condassign(Vout1,(x[16]-x[0]-vtp),0,Vout1);
    condassign(Idsn1,(x[16]-x[0]-vtp),lk,Idsn1);
    condassign(Idsp1,(x[16]-x[0]-vtp),-lk,Idsp1);
    
    // Region A2: 0 <= Vin2 <= Vtn
    Vout2 = x[0];
    Idsn2 = lk;
    Idsp2 = -lk;
   
    // Region B2: Vtn <= Vin2 <= Vdd/2 
    IdsnB2 = ((betan/2) * (x[17] - vtn) * (x[17] - vtn))+(power*x[0])+lk;
    IdspB2 = (-betap * (((x[17] - x[0] - vtp) * (x[3] - x[0])) 
       - ((x[3] - x[0]) * (x[3] - x[0]) / 2)))-(power*x[0])-lk;
    tempB2 = (x[17] - vtp)*(x[17] - vtp);
    tempB2 = tempB2 - (2 * x[0] * (x[17] - (x[0]/2) - vtp));
    tempB2 = tempB2 - ((betan/betap) * (x[17] - vtn) * (x[17] - vtn));
    VoutB2 = (x[17] - vtp) + sqrt(tempB2);
   
    condassign(Vout2,(x[17]-vtn),VoutB2,Vout2);
    condassign(Idsn2,(x[17]-vtn),IdsnB2,Idsn2);
    condassign(Idsp2,(x[17]-vtn),IdspB2,Idsp2);
     
    // Region C2: Vin2 = Vdd/2
    IdsnC2 = (betan * (x[17] - vtn) * (x[17] - vtn) / 2)+(power*x[0])+lk;
    IdspC2 = -IdsnC2;
    VoutC2 = x[3];
    
    // Region D2: Vdd/2 <= Vin2 <= Vdd + Vtp 
    IdsnD2 = (betan * (((x[17] - vtn) * x[3]) - (x[3] * x[3] / 2))) +
     (power*x[0])+lk ;
    IdspD2 = (- (betap/2) * (x[17] - x[0] - vtp) * (x[17] - x[0] - vtp))
      - (power*x[0])-lk ;
    tempD2 = (x[17] - vtn) * (x[17] - vtn);
    tempD2 = tempD2 - ((betap/betan) * (x[17] - x[0] - vtp) * (x[17] - x[0] - vtp));
    VoutD2 = (x[17] - vtn) - sqrt(tempD2);
    
    condassign(Vout2,(x[17]-(x[0]/2)),VoutD2,Vout2);
    condassign(Idsn2,(x[17]-(x[0]/2)),IdsnD2,Idsn2);
    condassign(Idsp2,(x[17]-(x[0]/2)),IdspD2,Idsp2);
    
    // Region E2: Vin2 >= Vdd + Vtp
    condassign(Vout2,(x[17]-x[0]-vtp),0,Vout2);
    condassign(Idsn2,(x[17]-x[0]-vtp),lk,Idsn2);
    condassign(Idsp2,(x[17]-x[0]-vtp),-lk,Idsp2);
    
    // Assign output voltage
    vp[0] = x[0];
    vp[1] = x[1];
    vp[2] = x[2];
    condassign(Vout,(x[17]-x[16]),Vout2,Vout1);
    vp[3] = Vout;
    vp[4] = tnom + x[4]*zt;

    // Assign output current
    condassign(Idsp,(x[17]-x[16]),Idsp2,Idsp1);
    condassign(Idsn,(x[17]-x[16]),Idsn2,Idsn1);
    ip[0] = -Idsp;
    ip[1] = 0;
    ip[2] = 0;
    ip[3] = Idsn1 + Idsn2 + Idsp;
    ip[4] = ((vp[0]-vp[3])*Idsp)-(vp[3]*(Idsn1+Idsn2))-(power*x[0])-lk;
}
