#include "../../../../network/CircuitManager.h"
#include "../../../../network/Element.h"
#include "../../../../network/AdolcElement2.h"
#include "DiodeCompact.h"

// Static members
const unsigned DiodeCompact::n_par = 18;

// Element information
ItemInfo DiodeCompact::einfo =
{
  "diodecompact",
  "Spice diode model (conserves charge)",
  "Carlos E. Christoffersen",
  DEFAULT_ADDRESS"category:diode",
  "2000_05_15"
};

// Parameter information
ParmInfo DiodeCompact::pinfo[] =
{
  {"js", "Saturation current (A)", TR_DOUBLE, false},
  {"alfa", "Slope factor of conduction current (1/V)", TR_DOUBLE, false},
  {"jb", "Breakdown saturation current (A)", TR_DOUBLE, false},
  {"vb", "Breakdown voltage (V)", TR_DOUBLE, false},
  {"e", "Power-law parameter of breakdown current", TR_DOUBLE, false},
  {"ct0", "Zero-bias depletion capacitance (F)", TR_DOUBLE, false},
  {"fi", "Built-in barrier potential (V)", TR_DOUBLE, false},
  {"gama", "Capacitance power-law parameter", TR_DOUBLE, false},
  {"r0", "Bias-dependent part of series resistance in forward-bias (Ohms)",
	TR_DOUBLE, false},
  {"t", "Intrinsic time constant of depletion layer (s)", TR_DOUBLE, false},
  {"area", "Area multiplier", TR_DOUBLE, false},
  {"imax", "Maximum forward and reverse current (A)", TR_DOUBLE, false},
  {"eg", "Barrier height at 0 K (eV)", TR_DOUBLE, false},
  {"m", "Grading coefficient", TR_DOUBLE, false},
  {"aro", "r0 linear temperature coefficient (1/K)", TR_DOUBLE, false},
  {"bro", "r0 quadratic temperature coefficient (1/K^2)", TR_DOUBLE, false},
  {"afag", "Temperature-related coefficient", TR_DOUBLE, false},
  {"xti", "Js temperature exponent", TR_DOUBLE, false}
};


DiodeCompact::DiodeCompact(const string& iname)
: AdolcElement2(&einfo, pinfo, n_par, iname)
{
  // Set default parameter values
  paramvalue[0] = &(js = 1e-16);
  paramvalue[1] = &(alfa = 38.696);
  paramvalue[2] = &(jb = 1e-5);
  paramvalue[3] = &(vb = -1e20);
  paramvalue[4] = &(e = 10.);
  paramvalue[5] = &(ct0 = zero);
  paramvalue[6] = &(fi = .8);
  paramvalue[7] = &(gama = .5);
  paramvalue[8] = &(r0 = zero);
  paramvalue[9] = &(t = zero);
  paramvalue[10] = &(area = one);
  paramvalue[11] = &(imax = zero);
  paramvalue[12] = &(eg = .8);
  paramvalue[13] = &(m = .5);
  paramvalue[14] = &(aro = zero);
  paramvalue[15] = &(bro = zero);
  paramvalue[16] = &(afag = one);
  paramvalue[17] = &(xti = 2.);

  // Set the number of terminals
  setNumTerms(2);
  // Set number of states
  setNumberOfStates(1);
  setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);
}

void DiodeCompact::init() throw(string&)
{
  DenseIntVector var(1,0);
  DenseIntVector novar;
  DenseDoubleVector nodelay;
  createTape1(var, novar, nodelay, 1, 2);
  createTape2();
}

void DiodeCompact::eval1(adoublev& x, adoublev& xt, adoublev& y1, adoublev& z1)
{
  double v1 = log(5e8 / alfa) / alfa; // normal is .5e9
  double k1 = ct0 / pow(.2, gama);
  double k2 = fi * .8;
  double k3 = exp(alfa * v1);

  // x[0]: state variable

  adouble vj, qj, rs, itmp;

  condassign(vj, v1 - x[0],
	x[0] + zero,
	v1 + log(one + alfa*(x[0]-v1))/alfa);

  // Calculate capacitance as usual
  if (isSet(&ct0))
    condassign(qj, vj - k2,
	ct0 * fi * (one - pow(.2, one - gama)) / (one - gama)
	+ k1 * fi * (vj / fi - .8),
	ct0 * fi * (one - pow(one - vj / fi, one - gama))
	/ (one - gama));

  // Now use the state variable again to calculate the total
	// current.  This way, we save some exp() calls.  The total
	// current is the current through the capacitor plus the ideal
	// diode current.
  condassign(itmp, v1 - x[0],
	js * (exp(alfa * x[0]) - one),
	js * k3 * (one + alfa * (x[0] - v1)) - js);
  z1[1] = itmp;

  // subtract the breakdown current
  condassign(itmp, vj - vb - one,
	zero ,
	jb * pow(one + vb - vj, e));
  z1[1] -= itmp;

  z1[0] = vj;
  y1[0] = qj;
}

void DiodeCompact::eval2(adoublev& dy1, adoublev& z1, adoublev& vp, adoublev& ip)
{
  double k1 = ct0 / pow(.2, gama);
  double k2 = fi * .8;
  adouble rs;
  // Calculate capacitance as usual
  if (isSet(&ct0))
  {
    adouble cj;
    condassign(cj, z1[0] - k2,  k1, ct0 / pow(one - z1[0]/fi, gama));
    condassign(rs, -r0 + t/cj, zero , r0 - t/cj);
  }
  else
    rs = r0;

  // Add capacitor current
  ip[0] = z1[1] + dy1[0];
  vp[0] = z1[0] + ip[0] * rs;

  // scale the current according to area.
  ip[0] *= area;
}

