#include "../../../../network/CircuitManager.h"
#include "../../../../network/Element.h"
#include "../../../../network/AdolcElement2.h"
#include "Mosnekvt.h"
#include "../../../../analysis/TimeDomainSV.h"

// Static members
const unsigned Mosnekvt::n_par = 44;

// Element information
ItemInfo Mosnekvt::einfo =
{
  "mosnekvt",
  "EPFL EKV MOSFET model",
  "Jin Liang, SiWei Chen and Michael McCall",
  DEFAULT_ADDRESS"elements/Mosnekvt.html",
  "2005_04_01"
};

// Parameter information
ParmInfo Mosnekvt::pinfo[] =
{
  {"type", "N-channel or P-channel MOS", TR_DOUBLE, false},
// Device input variables  
  {"l", "Gate length (m)", TR_DOUBLE, false},
  {"w", "Gate width (m)", TR_DOUBLE, false},
  {"np", "Parallel multiple device number", TR_DOUBLE, false},
  {"ns", "Serial multiple device number", TR_DOUBLE, false},
// Process related parameters
  {"cox", "Gate oxide capacitance per area (F/m^2)", TR_DOUBLE, false},
  {"xj", "Junction depth (m)", TR_DOUBLE, false}, 
  {"dw", "Channel width correction (m)", TR_DOUBLE, false},
  {"dl", "Channel length correction (m)", TR_DOUBLE, false}, 
// Basic intrinsic model parameters
  {"vto", "Long_channel threshold voltage (V)", TR_DOUBLE, false},
  {"gamma", "Body effect parameter (V^1/2)",TR_DOUBLE, false},
  {"phi", "Bulk Fermi potential (V)", TR_DOUBLE, false},
  {"kp", "Transconductance parameter (A/V^2)", TR_DOUBLE, false},
  {"eo", "Mobility reduction coefficient (V/m)",TR_DOUBLE, false},
  {"ucrit", "Longitudinal critical field (V/m)", TR_DOUBLE, false},
// Operational parameters 
  {"tox", "Oxide thickness (m)", TR_DOUBLE, false},
  {"nsub", "Channel doping (1/cm^3)", TR_DOUBLE, false},
  {"vfb", "Flat-band voltage (V)", TR_DOUBLE, false},
  {"uo", "Low-field mobility(cm^2/(V.s))", TR_DOUBLE, false},
  {"vmax", "Saturation velocity (m/s)", TR_DOUBLE, false},
  {"theta", "Mobility recuction coefficient (1/V)", TR_DOUBLE, false},
// Channel length modulation and charge sharing parameters
  {"lambda", "Channel-length modulation", TR_DOUBLE, false},
  {"weta", "Narrow-channel effect coefficient", TR_DOUBLE, false},
  {"leta", "Short-channel effect coefficient", TR_DOUBLE, false},
// Reverse short-channel effect parameters
  {"qo", "Reverse short channel effect peak charge density (A.s/m^2)", TR_DOUBLE, false},
  {"lk", "Reverse short channel effect characteristic length (m)", TR_DOUBLE, false},
// Impact ionization related parameters
  {"iba", "First impact ionization coefficient (1/m)", TR_DOUBLE, false},
  {"ibb", "Second impact ionization coefficient (V/m)", TR_DOUBLE, false},
  {"ibn", "Saturation voltage factor for impact ionization", TR_DOUBLE, false},
// Intrinsic model temperature parameters
  {"tcv", "Threshold voltage temperature coefficient (V/K)", TR_DOUBLE, false},
  {"bex", "Mobility temperature exponent", TR_DOUBLE, false},
  {"ucex", "Longitudinal critical field temperature exponent", TR_DOUBLE, false},
  {"ibbt", "Temperature coefficient for IBB (1/K)", TR_DOUBLE, false},
// Matching parameters
  {"avto", "Area related threshold voltage mismatch parameter (Vm)", TR_DOUBLE, false},
  {"akp", "Area related gain mismatch parameter (m)", TR_DOUBLE, false},
  {"agamma", "Area related body effect mismatch parameter (V^(1/2)m)", TR_DOUBLE, false},
// Flicker noise parameters
  {"kf", "Flicker noise coefficient", TR_DOUBLE, false},
  {"af", "Flicker noise exponent", TR_DOUBLE, false},
// Setup parameters
  {"nqs", "Non-Quasi-Static operation switch", TR_DOUBLE, false},  
  {"satlim", "Ratio defining the saturation limit if/ir", TR_DOUBLE, false},
  {"xqc", "Charge/capacitance model selector", TR_DOUBLE, false},
  
  {"scale", "Scale parameter", TR_DOUBLE, false},
  {"tnom", "Nominal temperature of model parameters (K)", TR_DOUBLE, false},
  {"tmp", "Model simulation temperature (K)", TR_DOUBLE, false}
};

Mosnekvt::Mosnekvt(const string& iname) : AdolcElement2(&einfo, pinfo, n_par, iname)
{
  // Set default parameter values
  paramvalue[0]  = &(type = NMOS);
  paramvalue[1]  = &(l = 1e-6);
  paramvalue[2]  = &(w = 1e-6);
  paramvalue[3]  = &(np = 1.0);  
  paramvalue[4]  = &(ns = 1.0);	
  paramvalue[5]  = &(cox = 0.0);
  paramvalue[6]  = &(xj = 1.0e-7);  
  paramvalue[7]  = &(dw = 0.0);
  paramvalue[8]  = &(dl = 0.0);
  paramvalue[9]  = &(vto = 0.0);
  paramvalue[10]  = &(gamma = 0.0);
  paramvalue[11]  = &(phi = 0.0);
  paramvalue[12] = &(kp = 0.0);   
  paramvalue[13] = &(eo = 0.0);   
  paramvalue[14] = &(ucrit = 0.0); 
  paramvalue[15] = &(tox = 0.0);
  paramvalue[16] = &(nsub = 0.0);
  paramvalue[17] = &(vfb = -2003);
  paramvalue[18] = &(uo = 0.0);
  paramvalue[19] = &(vmax = 0.0);
  paramvalue[20] = &(theta = -1); 
  paramvalue[21] = &(lambda = 0.5); 
  paramvalue[22] = &(weta = 0.25);  
  paramvalue[23] = &(leta = 0.1);   
  paramvalue[24] = &(qo = 0.0);   
  paramvalue[25] = &(lk = 2.9e-7); 
  paramvalue[26] = &(iba = 0.0); 
  paramvalue[27] = &(ibb = 3.0e8);   
  paramvalue[28] = &(ibn = 1.0);
  paramvalue[29] = &(tcv = 1.0e-3);
  paramvalue[30] = &(bex = -1.5);
  paramvalue[31] = &(ucex = 0.8);
  paramvalue[32] = &(ibbt = 9.0e-4);  
  paramvalue[33] = &(avto = 0.0);
  paramvalue[34] = &(akp = 0.0);
  paramvalue[35] = &(agamma = 0.0);
  paramvalue[36] = &(kf = 0.0);
  paramvalue[37] = &(af = 1.0);
  paramvalue[38] = &(nqs = 0.0);
  paramvalue[39] = &(satlim = 54.59815003314424);
  paramvalue[40] = &(xqc = 0.4);
  paramvalue[41] = &(scale = 1);
  paramvalue[42] = &(tnom = 300.15);
  paramvalue[43] = &(tmp = 300.15);
  //new info
 // paramvalue[44] = &(hf = 0.0);

  // Set the number of terminals
  //setNumTerms(4);
  setNumTerms(6); //two extra terminals from thermal resistor

  // Set flags
  //setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);
  setFlags(NONLINEAR|MULTI_REF|TR_TIME_DOMAIN);

  // Set number of states
  // vgs,vds,vsb,Qg,Qd, Qb
  setNumberOfStates(4); //added tmp

    // allocate the voltage and current vectors
  double lowerDim = -40.0; 
  double upperDim = 4.0; 
  double points = 5000.0;
  double step = (upperDim - lowerDim)/points;
  length = (int)points;

  // create voltages and currents
  voltages = new double[length];
  currents = new double[length];

  double powerTerm, r, a, b;

  // first generate a logspace as in MATLAB
  // then fill voltage and current vectors with function values
  for (int ii = 0; ii < length; ii++)
  {
    powerTerm = lowerDim + ii*(upperDim - lowerDim)/(points - 1.0);
    currents[ii] = pow(10.0, powerTerm);
    r = currents[ii] + 0.5*sqrt(currents[ii]) + 1.0;
    a = log((sqrt(currents[ii]) + 4.0*(sqrt(currents[ii]) + 1.0))/8.0);
    b = log((1.0 + 4.0*(sqrt(currents[ii]) + sqrt(r)))/5.0);
    voltages[ii] = log(currents[ii]) + 2.0*(sqrt(r) - 1.0) - 2.0*a + 0.5*b;
  }
}

Mosnekvt::~Mosnekvt()
{
  delete [] currents;
  delete [] voltages;
}

void Mosnekvt::init() throw(string&)
{
  // Set Constants
  epsilonox = scale*34.5e-12;
  epsilonsi = scale*104.5e-12;
  q=1.602e-19;
  k=1.3807e-23;
  Tref=300.15;    
  vtTnom = (k*tnom)/q;		
  vtTref = (k*Tref)/q;
  egTnom = 1.16-0.000702*tnom*tnom/(tnom+1108.0);
  egTref = 1.16-0.000702*Tref*Tref/(Tref+1108.0);
  niTnom = 1.45e16*(tnom/Tref)*exp(egTref/(2.0*vtTref)-egTnom/(2.0*vtTnom));

  // For P-channel devices
  if(type == -1)
  {
    tcv = type * tcv;
    vto = type * vto;
  }

  // Calculate any missing parameters from user-defined settings
  // COX 
  if(cox == 0.0)
  {
    if (tox > 0.0)
      cox = epsilonox / tox;
    else
      cox = 0.7e-3;
  }

  // GAMMA
  if(gamma == 0.0)
  {
    if (nsub > 0.0)
      gamma = sqrt(2.0 * epsilonsi * nsub * 1e6) / cox;
    else
      gamma = 1.0;
  }

  // PHI
  if(phi == 0.0)
  {
    if (nsub > 0.0)
      phi = 2.0 * vtTnom * log(nsub * 1e6 / niTnom);
    else
      phi = 0.7;
  }

  // VTO
  if(vto == 0.0)
  {
    if(vfb != -2003.0)
      vto = vfb + phi + gamma * sqrt(phi);
    else
      vto = 0.5;
  }

  // KP
  if(kp == 0.0)
  {
    if(uo > 0.0) 
      kp = uo * cox * 1e-4 /*(m^2/cm^2)*/;
    else
      kp = 5.0e-5;
  }

  // UCRIT
  if(ucrit == 0.0)
  {
    if((vmax > 0) && (uo > 0))
      ucrit = vmax / (uo * 1e-4);
    else
      ucrit = 2.0e6;
  }

  // EO
  if(eo == 0.0)
  {
    if(theta >=0)
      eo = 0.0;
    else
      eo = 1e12;
  }

  // create tape  
  DenseIntVector var(4);
  var[0] = 0;
  var[1] = 1;
  var[2] = 2;
  var[3] = 3; //tmp
  DenseIntVector novar;
  DenseDoubleVector nodelay;
  createTape1(var, novar, nodelay, 4, 8);
  createTape2();
}

void Mosnekvt::getLocalRefIdx(UnsignedVector& local_ref_vec,TerminalVector& term_list)
{
  // Make sure the vectors are empty
  term_list.erase(term_list.begin(), term_list.end());
  local_ref_vec.erase(local_ref_vec.begin(), local_ref_vec.end());

  // Insert vector elements
  term_list.push_back(getTerminal(0));	//drain
  term_list.push_back(getTerminal(1));	//gate
  term_list.push_back(getTerminal(2));	//source
  term_list.push_back(getTerminal(3)); 	// Local reference terminal		bulk
  term_list.push_back(getTerminal(4));
  term_list.push_back(getTerminal(5)); // Local reference terminal
  local_ref_vec.push_back(3); // Local reference index
  local_ref_vec.push_back(5); // Local reference index
}

void Mosnekvt::eval1(adoublev& x, adoublev& xt, adoublev& y1, adoublev& z1)
{
  // x[0]: vds
  // x[1]: vgs
  // x[2]: vbs
  // x[3]: tmp - tnom

  // x[4]: dvds/dt
  // x[5]: dvgs/dt
  // x[6]: dvbs/dt
  // x[7]: dtmp/dt

  // z1[3]: vdb , z2[3]: id
  // z1[4]: vgb , z2[4]: ig
  // z1[5]: vsb , z2[5]: is  
  // z1[6]: tmp , z2[7]: hf		//heat current

  adouble vtoT, kpT, ucritT, phiT, ibbT;
  adouble vtT,egT,niT;

  z1[7] = x[3]+tnom;		//tmp
 // z1[7] = x[3] + tmp;

  vtT = (k*z1[7])/q;	// Thermal voltage  
  egT = 1.16-0.000702*z1[7]*z1[7]/(z1[7]+1108.0);		// Energy gap
  niT = 1.45e16*(z1[7]/Tref)*exp(egTref/(2.0*vtTref)-egT/(2.0*vtT));  

  // Intrinsic parameters temperature dependence
  vtoT=vto-tcv*(z1[7]-tnom);
  kpT=kp*pow(z1[7]/tnom,bex);
  ucritT=ucrit*pow(z1[7]/tnom,ucex);
  phiT=phi*z1[7]/tnom-3*vtT*log(z1[7]/tnom)-egTnom*z1[7]/tnom+egT;
  ibbT=ibb*(1.0+ibbt*(z1[7]-tnom));

  double weff = w + dw;
  double leff = l + dl;
  adouble vtoa = vtoT + avto / sqrt(np * weff * ns * leff);
  adouble kpa = kpT * (1.0 + akp / sqrt(np * weff * ns * leff));
  double gammaa = gamma + agamma / sqrt(np * weff * ns * leff);
  double cEps = 4.0 * pow(22.0e-3,2);
  double cA = 0.028;
  double xi  = cA * (10.0 * leff / lk - 1.0);
  double deltavRSCE = 2.0 * qo / (cox * pow(1.0 + 0.5 * (xi + sqrt(xi*xi + cEps)), 2));
  adouble vc = ucritT * ns * leff;
  double lc = sqrt(epsilonsi * xj / cox);
  double lmin = ns * leff / 10.0;
  if(type == 1)
    eta = 0.5;
  else
    eta = 0.3333333333333;
  adouble qbo = gammaa * sqrt(phiT);
  double qox = 0.0;
  double C_ox = cox * np * weff * ns * leff;

  //All the active variables for ADOL-C "adoubles" must be initialiazed here
  adouble vgprime, vpo1, vpo, vsprime, vdprime, gammao, gammaprime, vp1, vp;
  adouble n, i_f, vdss, vdssprime, deltav, vds, vip, deltal, lprime, leq;
  adouble irprime, ir, betao, betaoprime, beta, is, ids, vib; //vpprime;
  adouble idb1, idb, id, nq, xf, xr, qd, qs, qi, qb1, qb, qg, QI, QB, QD;
  adouble QS, QG; 

  // Effective gate voltage in cluding reverse short channel effect
  vgprime =  type*x[1] - vtoa - deltavRSCE + phiT + gammaa * sqrt(phiT);

  // Effective substrate factor including charge-sharing for short and narrow
  // channels
  // Pinch-off voltage for narrow-channel effect 
  vpo1 = vgprime - phiT - gammaa * (sqrt(vgprime + gammaa*gammaa / 4.0) - gammaa
  / 2);
  condassign(vpo, vgprime, vpo1, -phiT);

  // Effective substrate factor accounting for charge-sharing
  vsprime = 0.5 * (type*x[2] + phiT + sqrt(pow(type*x[2] + phiT,2) + 16.0 * vtT*vtT));
  vdprime = 0.5 * (type*x[0] + phiT + sqrt(pow(type*x[0] + phiT,2) + 16.0 * vtT*vtT));

  // Pinch-off voltage including short- and narrow-channel effect
  gammao = gammaa - epsilonsi * (leta * (sqrt(vsprime) + sqrt(vdprime))
  / leff - 3.0 * weta * sqrt(vpo + phiT) / weff) / cox;
  gammaprime = 0.5 * (gammao + sqrt(gammao*gammao + 0.1 * vtT));
  vp1 = vgprime - phiT - gammaprime * (sqrt(vgprime+pow(gammaprime/2.0, 2)) -
  gammaprime/2);
  condassign(vp, vgprime, vp1, -phiT);

  // Slope factor
  n = 1.0 + gammaa / (2.0 * sqrt(vp + phiT + 4.0 * vtT));

  // Forward normalized current
  //i_f = log(1 + exp((vp - type*x[2]) / (2 * vtT)))*log(1 + exp((vp - type*x[2]) / (2 * vtT)));
  adouble v1 = (vp - type*x[2])/vtT;
  i_f = linearInterpolate(v1.value());

  // Velocity saturation voltage
  vdss = vc * (sqrt(0.25 + vtT * sqrt(i_f) / vc) - 0.5);

  // Drain-to-source saturation voltage for reverse normalized current
  vdssprime = vc * (sqrt(0.25 + vtT * (sqrt(i_f) - 0.75 * log(i_f))/vc) - 0.5) +
  vtT * (log(vc / (2 * vtT)) - 0.6);

  // Channel-length modulation
  deltav = 4.0 * vtT * sqrt(lambda * (sqrt(i_f) - vdss / vtT) + 1.0/64.0);
  vds = (type*x[0] - type*x[2])/2.0;
  vip = sqrt(vdss*vdss + deltav*deltav) - sqrt(pow(vds - vdss,2) + deltav*deltav);
  deltal = lambda * lc * log(1.0 + (vds - vip) / (lc * ucritT));

  // Equivalent channel length including channel-length moculation and velocity
  // saturation
  lprime = ns * leff - deltal + (vds + vip) / ucritT;
  leq = 0.5 * (lprime + sqrt(lprime*lprime + lmin*lmin));

  // Reverse normalized current
  //irprime = log(1 + exp(((vp - vds - type*x[2] - sqrt(vdssprime*vdssprime +
  //deltav*deltav) + sqrt((vds-vdssprime)*(vds-vdssprime) + deltav*deltav)) /
  //vtT)/ 2))*log(1 + exp(((vp - vds - type*x[2] - sqrt(vdssprime*vdssprime +
  //deltav*deltav) + sqrt((vds-vdssprime)*(vds-vdssprime) + deltav*deltav)) / vtT)/ 2));
  adouble v2 = (vp - vds - type*x[2] - sqrt(pow(vdssprime,2) + pow(deltav,2)) + sqrt(pow(vds - vdssprime, 2) + pow(deltav, 2)))/vtT;
  irprime = linearInterpolate(v2.value());

  // Reverse normalized currect for mobility model, intrinsic
  //charges/capacitances, thermal noise model and NQS time-constant
  //ir = log(1 + exp((vp -  type*x[0]) / (2 * vtT)))*log(1 + exp((vp -  type*x[0]) / (2 * vtT)));
  adouble v3 = (vp - type*x[0])/vtT;
  ir = linearInterpolate(v3.value());

  // Transconductance factor and mobility reduction due to vertical field
  betao = kpa * np * weff / leq;
  betaoprime = betao * (1.0 + cox * qbo / (eo * epsilonsi));

  // Quasi-static model equations
  // Dynamic model for the intrinsic node charges
  nq = 1.0 + gammaa / (2.0 * sqrt(vp + phiT + 1.0e-6));

  // Normalized intrinsic node charges
  xf = sqrt(0.25 + i_f);
  xr = sqrt(0.25 + ir);
  qd = -nq * (4.0 * (3.0*xr*xr*xr + 6.0*xr*xr*xf + 4.0*xr*xf*xf + 2*xf*xf*xf) / (15.0 *
  pow(xf + xr,2)) - 0.5);
  qs = -nq * (4.0 * (3.0*xf*xf*xf + 6.0*xf*xf*xr + 4.0*xf*xr*xr + 2.0*xr*xr*xr) / (15.0*
  pow(xf + xr,2)) - 0.5);
  qi = qs + qd;
  qb1 = -gammaa * sqrt(vp + phiT + 1.0e-6) / vtT - (nq - 1.0) * qi / nq;
  condassign(qb, vgprime, qb1, -vgprime / vtT);
  qg = -qi - qox - qb;

  // Rigorous mobility reduction model
  beta = betaoprime / (1.0 + cox * vtT * fabs(qb + eta*qi) / (eo * epsilonsi));

  // Specific current
  is = 2.0*n*beta*vtT*vtT;

  // Drain-to-source current
  ids = type*is*(i_f - irprime);

  // Impact ionization current
  vib =  type*x[0] - type*x[2] - 2.0*ibn*vdss;
  idb1 = ids * iba * vib * exp(-ibbT * lc / vib) / ibbT;
  condassign(idb, vib, idb1, 0.0);
  id = ids + idb; 

  // Total charges
  QI = C_ox * vtT * qi;
  QB = C_ox * vtT * qb;
  QD = C_ox * vtT * qd;
  QS = C_ox * vtT * qs;
  QG = C_ox * vtT * qg;

  y1[0] = type*(QD-QB);
  y1[1] = type*(QG-QB);
  y1[2] = type*(QS-QB);

  // Assign DC currents
  z1[0] =  id;    //DC Drain current
  z1[1] = 0.0; 		//DC Gate current
  z1[2] = -id; 	//DC Source current
  z1[3] = -type*id*vds;
  //z1[3] = ids;

  // Assign known output voltages
  z1[4] = x[0];  // vdb
  z1[5] = x[1];	 // vgb 
  z1[6] = x[2];	 // vsb 

  //z1[3] = -z1[0]*z1[4];
}

void Mosnekvt::eval2(adoublev& dy1, adoublev& z1, adoublev& vp, adoublev& ip)
{
  vp[0] = z1[4]; //Vdb
  vp[1] = z1[5]; //Vgb
  vp[2] = z1[6]; //Vsb

  vp[3] = z1[7];	//tmp

  ip[0] = z1[0] + dy1[0]; //Drain current  Drain charge derivative
  ip[1] = z1[1] + dy1[1]; //Gate current, gate charge derivative
  ip[2] = z1[2] - dy1[2]; //Source current, source charge derivative

  ip[3] = z1[3];
}


// search for the correct index in the voltage array
// based on the GNU GSL library interpolation function
int Mosnekvt::bSearch(double xval, int index_lo, int index_hi)
{
  int ilo = index_lo;
  int ihi = index_hi;
  while (ihi > ilo + 1)
  {
    int i = (ihi + ilo) / 2;
    if (voltages[i] > xval)
      ihi = i;
    else
      ilo = i;
  }

  return ilo;
}

int Mosnekvt::accelFind(double xval, int len)
{
  int x_index = 0;
  int idx;

  if (xval < voltages[x_index])
  {
    idx = bSearch (xval, 0, x_index);
  }
  else if (xval > voltages[x_index + 1])
  {
    idx = bSearch (xval, x_index, len - 1);
  }

  return idx;
}

// return interpolated value
// based on the GNU GSL library interpolation function
adouble Mosnekvt::linearInterpolate(const adouble xval)
{
  // xval is used to get the correct index
  // from the voltage array
  // get the array index using a search function
  int index = bSearch(xval.value(), 0, length-1);
  //int index = accelFind(xval.value(), length);

  adouble yval;

  double x_lo = voltages[index];
  double x_hi = voltages[index + 1];
  double y_lo = currents[index];
  double y_hi = currents[index + 1];
  double dx = x_hi - x_lo;
  if (dx > 0.0)
    yval = y_lo + (xval - x_lo) / dx * (y_hi - y_lo);
  else
    yval = 0.0;

  return yval;

}
