// Based on the Lueder model
#include "../../../../network/CircuitManager.h"
#include "../../../../network/Element.h"
#include "../../../../network/AdolcElement2.h"
#include "Mosntftq.h"
#include "../../../../analysis/TimeDomainSV.h"

// Static members
const unsigned Mosntftq::n_par = 16;
const double E0 = 8.854194e-12;
const double q = 1.6e-19;
int counter=0;
int flag=0;

// Element information
ItemInfo Mosntftq::einfo =
{
  "mosntftq",
  "a-Si TFT MODEl",
  "ECe718 Class Project",
  DEFAULT_ADDRESS"transistor>mosfet",
  "2003_05_15"
};

// Parameter information
ParmInfo Mosntftq::pinfo[] =
{
  {"l", "Channel length (m)",TR_DOUBLE, false},
  {"w", "Channel width (m)", TR_DOUBLE, false},
  {"u", "Surface mobility (cm^2/V-s)", TR_DOUBLE, false},
  {"vto", "Zero-bias threshold voltage (V)", TR_DOUBLE, false},
  {"tts", "Device temperature (K)",TR_DOUBLE,false},
  {"ctr", "Mobility modulation (V^-1)", TR_DOUBLE, false},
  {"rd", "Static feedback on threshold voltage (V^-1) ", TR_DOUBLE, false},
  {"vthm", "Saturated velocity (m/sec)", TR_DOUBLE, false},
  {"beta", "Conductance (ohm-1)", TR_DOUBLE, false},
  {"roff", "Coefficient of drain leakage", TR_DOUBLE, false},
  {"cgso", "Gate-source capacitance (F)", TR_DOUBLE, false},
  {"cgdo", "Gate-drain capacitance (F)", TR_DOUBLE, false},
  {"cl", "Maximum space-charge capacitance (F/m^2)", TR_DOUBLE, false},
  {"lambda", "signal frequency (Hz)",TR_DOUBLE, false},
  {"rs", "Frequenct effect compensation", TR_DOUBLE, false},
  {"tnom", "Frequenct effect compensation", TR_DOUBLE, false}
};

Mosntftq::Mosntftq(const string& iname)
: AdolcElement2(&einfo, pinfo, n_par, iname)
{
  // Set default parameter values
  paramvalue[0] = &(l = 5.5e-6);
  paramvalue[1] = &(w = 80e-6);
  paramvalue[2] = &(u = 0.12e-4);
  paramvalue[3] = &(vto = 3.2);
  paramvalue[4] = &(tts = 232);
  paramvalue[5] = &(ctr = 0.06);
  paramvalue[6] = &(rd = 83e3);
  paramvalue[7] = &(vthm = 4.2 );
  paramvalue[8] = &(beta = 3);
  paramvalue[9] = &(roff = 4e15);
  paramvalue[10] = &(cgso = 80e-15);
  paramvalue[11] = &(cgdo = 80e-15);
  paramvalue[12] = &(cl = 400e-6);
  paramvalue[13] = &(lambda = 450);
  paramvalue[14] = &(rs= 83e3);
  paramvalue[15] = &(tnom= 300.15);

  // Set the number of terminals
  setNumTerms(4);

  // Set flags
  setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);

  // Set number of states
	setNumberOfStates(3);
}

void Mosntftq::init() throw(string&)
{
  // constants
  A = exp(-(((log(1/ctr))/(vthm-vto))*vto));
  D = ((log(1/ctr))/(vthm-vto));

  // vt
	vt = (kBoltzman * tnom) / q;

  // create tape
  DenseIntVector var(3);
  var[0] = 0;
  var[1] = 1;
  var[2] = 2;

  DenseIntVector novar;
  DenseIntVector tvar;
  DenseDoubleVector nodelay;
  createTape1(var,novar,nodelay,3,6);
  createTape2();
}

void Mosntftq::eval1(adoublev& x, adoublev& xt, adoublev& y1, adoublev& z1)
{
  // x[0]: vds
  // x[1]: vgs
  // x[2]: vds - vgs
  // x[3]: dvds/dt
  // x[4]: dvgs/dt
  // x[5]: d(vgs - vds) / dt
  // vp[0]: ugs , ip[0]: id
  // vp[1]: uds , ip[1]: ig
  // vp[2]: usb , ip[2]:is
  adouble MODE;

  adouble  Ids;
  adouble vgs,vds,vbs,qdrn,qgate,qsrc;
  adouble kp,vtheff,kt,xs;

  vtheff = vthm-vto;
  kp = (w/l)*u*cl;
  kt = -4.0/15.0*pow((tnom/tts),2.0)+tnom/tts+1.0/15.0;
  xs = (pow((1-x[0]/lambda),-1))/2.0;

	//Subtreshold region
  if(x[1]<vto)
	{

		if(x[0]<vtheff)
			Ids = kp *(pow(vtheff,2.0/kt) - pow((vtheff-x[0]),2.0/kt)) * exp(beta*(x[1]-vto)) * xs * ctr;

		if(x[0]>vtheff)
			Ids = kp * (pow(vtheff, 2.0/kt)) * exp(beta*(x[1]-vto)) * xs * ctr;
	}

  //Transition region
  if((x[1]>=vto)&&(x[1]<vthm))
	{
		if(x[0]<vtheff)
			Ids = kp *(pow(vtheff,2.0/kt) - pow((vtheff-x[0]),2.0/kt)) * exp(D*x[1]) * xs * ctr * A;
		if(x[0]>vtheff)
			Ids = kp *(pow(vtheff,2.0/kt)) * exp(D*x[1]) * xs * ctr * A;
	}

  //above threshold region
  if(x[1]>=vthm)
	{
		if(x[0]<(x[1]-vto))
			Ids = kp *(pow((x[1]-vto),2.0/kt) - pow((x[1]-vto-x[0]),2.0/kt)) * xs ;
		if(x[0]>vtheff)
			Ids = kp *(pow((x[1]-vto),2.0/kt)) * xs ;
	}

  adouble ck, a;
  a= (x[1]-vto)/(x[0]-vto);
  ck = w*l*cl;

  qdrn =0;
  qgate=0;
  qsrc=0;
  adouble temp=0;
  temp= 1.0+2.0/kt;
  if((a>=0) && (a!=1))
	{
		qgate = ck*(x[1]-vto)*(1.0-pow(a,temp));
		qgate = qgate/((1.0+kt/2.0)*(1-pow(a,2.0/kt)));
		temp = 1.0+4.0/kt;

		qdrn = -qgate/(1-pow(a,2.0/kt)) + ck*(x[1]-vto)*(1.0-pow(a,temp))/((2.0+kt/2.0)*(1-pow(a,2.0/kt))*(1-pow(a,2.0/kt)));

		qsrc = qgate*pow(a,2.0/kt)/(1-pow(a,2.0/kt)) -ck*(x[1]-vto)*(1.0-pow(a,temp))/((2.0+kt/2.0)*(1-pow(a,2.0/kt))*(1-pow(a,2.0/kt)));
	}

  y1[0] =  qdrn;
  y1[1] =  qgate;
  y1[2] =  qsrc;

  z1[0] =  Ids; //DC Drain current
  z1[1] = 0.0; //DC Gate current
  z1[2] =  -Ids; //DC Source current

  z1[3] = x[0];  //Vdb
  z1[4] = x[1]; //Vgb
  z1[5] = x[2];//Vsb
}

void Mosntftq::eval2(adoublev& dy1, adoublev& z1, adoublev& vp, adoublev& ip)
{
  vp[0] = z1[3]; //Vdb
  vp[1] = z1[4]; //Vgb
  vp[2] = z1[5]; //Vsb

  ip[3] = z1[0] + dy1[0]; //Drain current  Drain charge derivative
  ip[4] = z1[1] + dy1[1]; //Gate current, gate charge derivative
  ip[5] = z1[2] - dy1[2]; //Source current, source charge derivative
}

