#include "../../../../network/ElementManager.h"
#include "../../../../network/AdolcElement2.h"
#include "Mosp3.h"

// Static members
const unsigned Mosp3::n_par = 22;

// Element information
ItemInfo Mosp3::einfo =
{
  "mosp3",
  "Intrinsic Mosfet model level 3",
  "Nikhil Kriplani",
  DEFAULT_ADDRESS"transistor>mosfet",
  "2002_10_10"
};

// Parameter information
ParmInfo Mosp3::pinfo[] =
{
  {"GAMMA", "Bulk threshold parameter (V^1/2)",TR_DOUBLE, false},
  {"KP", "transconductance parameter (A/V^2)", TR_DOUBLE, false},
  {"LD", "Lateral diffusion (m)", TR_DOUBLE, false},
  {"NSUB", "Substrate doping (1/m^3)", TR_DOUBLE, false},
  {"PHI", "Surface inversion potential (V)", TR_DOUBLE, false},
  {"TOX", "Oxide thickness (m)", TR_DOUBLE, false},
  {"U0", "Surface mobility (cm^2/V-s)", TR_DOUBLE, false},
  {"VT0", "Zero-bias threshold voltage (V)", TR_DOUBLE, false},
  {"WD", "Lateral diffusion width (m)", TR_DOUBLE, false},
  {"W", "Channel width (m)", TR_DOUBLE, false},
  {"L", "Channel length (m)", TR_DOUBLE, false},
  {"T", "Temperature of device (K)", TR_DOUBLE, false},
  {"TNOM", "Nominal temperature (K)", TR_DOUBLE, false},
  {"NFS", "Fast surface state density (cm^-2)", TR_DOUBLE, false},
  {"ETA", "Static feedback on threshold voltage ", TR_DOUBLE, false},
  {"THETA", "Mobility modulation (1/V)", TR_DOUBLE, false},
  {"VMAX", "Maximum drift velocity of carriers (m/sec)", TR_DOUBLE, false},
  {"XJ", "Metallurgical junction depth (m)", TR_DOUBLE, false},
  {"DELTA", "Width effect on threshold voltage", TR_DOUBLE, false},
  {"NSS", "Surface state density (1/cm^2)", TR_DOUBLE, false},
  {"TPG", "Type of gate material",TR_DOUBLE, false},
  {"KAPPA", "Saturation field factor", TR_DOUBLE, false}
};


Mosp3::Mosp3(const string& iname)
: AdolcElement2(&einfo, pinfo, n_par, iname)
{
  // Set default parameter values
  paramvalue[0] = &(GAMMA = 0.0);
  paramvalue[1] = &(KP = 0.000021);
  paramvalue[2] = &(LD = 0.0);
  paramvalue[3] = &(NSUB = 0.0);//was 1e15
  paramvalue[4] = &(PHI = 0.6);
  paramvalue[5] = &(TOX = 1.0e-7);
  paramvalue[6] = &(U0 = 600.0);
  paramvalue[7] = &(VT0 = 0.0);
  paramvalue[8] = &(WD = zero);
  paramvalue[9] = &(W = 50.0e-6);
  paramvalue[10] = &(L = 2.0e-6);
  paramvalue[11] = &(T = 300.15);
  paramvalue[12] = &(TNOM = 300.15);
  paramvalue[13] = &(NFS = 0.0);
  paramvalue[14] = &(ETA = 0.0);
  paramvalue[15] = &(THETA = 0.0);
  paramvalue[16] = &(VMAX = 0.0);
  paramvalue[17] = &(XJ = 0.0);
  paramvalue[18] = &(DELTA = 0.0);
  paramvalue[19] = &(NSS = 0.0);
  paramvalue[20] = &(TPG = 0.0);
  paramvalue[21] = &(KAPPA = 0.2);

  // Set the number of terminals
  setNumTerms(4);

  // Set flags
  setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);

  // Set number of states
  setNumberOfStates(3);
}


void Mosp3::init() throw(string&)
{
  pi = 3.14159265358979323846;
  epsilon0 = 8.854194e-12;
  epsilons = 11.7 * epsilon0;
  leff = L - 2.0 * LD;
  weff = W - 2.0 * WD;
  vt = (kBoltzman * T)/eCharge;
  cox = (3.9 * epsilon0) / TOX;
  oxidecap = cox * W * leff;

  // create tape
  DenseIntVector var(3);
  var[0] = 0;
  var[1] = 1;
  var[2] = 2;
  DenseIntVector novar;
  DenseDoubleVector nodelay;
  createTape1(var, novar, nodelay, 3, 6);
  createTape2();
}

void Mosp3::eval1(adoublev& x, adoublev& xt, adoublev& y1, adoublev& z1)
{
  // x[0]: vds
  // x[1]: vgs
  // x[2]: vbs
  // x[3]: dvds/dt
  // x[4]: dvgs/dt
  // x[5]: dvbs/dt
  // vp[0]: vdb , ip[0]: id
  // vp[1]: vgb , ip[1]: ig
  // vp[2]: vsb , ip[2]: is

  adouble MODE, T1, T2;

  if (x[0] >= 0.0)
    MODE = 1.0; //forward mode of operation
  else
    MODE = -1.0; //inverse mode of operation

  double eta = ETA * 8.15e-22 / (cox * leff * leff * leff);
  double ratio = T / TNOM;
  double ratio4 = ratio * sqrt(ratio);
  double fact2 = ratio; //need to check on this later, not sure of the difference
  double egfet = 1.16 - (7.02e-4 * T * T) / (T + 1108.0);
  double argt = -egfet / (kBoltzman*T*2.0) + 1.1150877 / (kBoltzman*(TNOM+TNOM));
  double pbfact = -2.0 * vt * (1.5*log(fact2) + eCharge * argt);
  double fact1 = 1.0; //TNOM/TREF
  double vtnom = kBoltzman * TNOM / eCharge;
  double egfet1 = 1.16 - (7.02e-4 * TNOM * TNOM) / (TNOM + 1108.0);
  double arg1 = -egfet1 / (kBoltzman*TNOM*2.0) + 1.1150877 / (kBoltzman*(TNOM+TNOM));
  double pbfact1 = -2.0 * vtnom * (1.5*log(fact1) + eCharge * arg1);
  double phio = (PHI - pbfact1) / fact1;
  double tPhi = fact2 * phio + pbfact;
  double TYPE = -1.0;

  double alpha;
  if (NSUB != 0.0)
    alpha = (2.0 * epsilons)/(eCharge * NSUB * 1.0e6);
  else
    alpha = 0.0;

  //all the adoubles must be initialiazed here
  adouble vdsat = 0.0;
  adouble phibs, sqphbs, sqphis, sqphs3, vbix, vth, von, cdl, fd2;
  adouble fshort, argb, argc, arga, wconxj, wponxj, djonxj, wps, onxn;
  adouble gammas, fbodys, fbody, onfbdy, qbonco, cdo, cdnorm, beta;
  adouble csonco, cdonco, xn, Ids, vgsx, vdsx, scf, wicf, wfact;
  adouble onfg, fgate, us, ondvt, delxl, vdsc, onvdsc, fdrain;
  adouble dlonxl, xlfact, cdsat, emax, gdsat, Ires;

  beta = (KP/ratio4) * (W/leff);

  if (MODE == 1.0)
    T1 = x[2];
  else
    T1 = TYPE * x[2] - TYPE * x[0];

  //square root term
	if(T1 <= 0.0)
	{
		phibs = tPhi - T1;
		sqphbs = sqrt(phibs);
	}
  else
	{
		sqphis = sqrt(tPhi);
		sqphs3 = tPhi * sqphis;
		sqphbs = sqphis / (1.0 + T1/(tPhi + tPhi));
		phibs = sqphbs * sqphbs;
	}

  double coeff0 = 0.0631353;
  double coeff1 = 0.8013292;
  double coeff2 = -0.01110777;

  //short channel effect factor
  if((XJ != 0.0) && (sqrt(alpha) != 0.0))
  {
    wps = sqrt(alpha) * sqphbs;
    djonxj = LD / XJ;
    wponxj = wps / XJ;
    wconxj = coeff0 + coeff1*wponxj + coeff2*wponxj*wponxj;
    arga = wconxj + djonxj;
    argc = wponxj / (1.0 + wponxj);
    argb = sqrt(1.0 - argc*argc);
    fshort = 1.0 - (XJ/leff) * (arga*argb - djonxj);
  }
  else
		fshort = 1.0;

  //Body Effect
  double narrowFactor;
  gammas = GAMMA * fshort;
  fbodys = 0.5 * gammas/(sqphbs + sqphbs);
  narrowFactor = DELTA * 0.5 * pi * epsilons/cox;
  fbody = fbodys + narrowFactor/W;
  onfbdy = 1.0 / (1.0 + fbody);
  qbonco = gammas * sqphbs + narrowFactor * phibs/W;

  double tVbi = VT0 - GAMMA*sqrt(PHI) + 0.5*(egfet1 - egfet) + 0.5*(tPhi - PHI);

  //threshold voltage
  vbix = tVbi * TYPE - eta * (MODE * x[0]);
  vth = vbix + qbonco;
  von = vth;

  if (MODE == 1.0)
    T1 = TYPE * x[1];
  else
    T1 = TYPE * x[1] - TYPE * x[0];

  if (NFS != 0.0)
  {
    csonco = eCharge * NFS * 1.0e4 * leff * W / oxidecap;
    cdonco = qbonco / (phibs + phibs);
    xn = 1.0 + csonco + cdonco;
    von = vth + vt*xn;
  }
  else //cutoff region
  {
    if (T1 <= von)
    {
      Ids = 0.0;
      goto the_end;
    }
  }

  condassign(vgsx, T1 - von, T1, von);

  //saturation voltage
  //Mobility modulation by gate voltage
  onfg = 1.0 + THETA * (vgsx - vth);
  fgate = 1.0 / onfg;
  us = U0/ratio4 * 1.0e-4 * fgate;

  // Saturation voltage
  vdsat = (vgsx - vth) * onfbdy;
  if (VMAX > 0.0)
  {
    vdsc = leff * VMAX / us;
    onvdsc = 1.0 / vdsc;
    arga = (vgsx - vth) * onfbdy;
    argb = sqrt(arga*arga + vdsc*vdsc);
    vdsat = arga + vdsc - argb;
  }

  //Linear region
  condassign(vdsx, vdsat - MODE * x[0], MODE * x[0] + zero, vdsat);

  if (vdsx == 0.0) goto line900; //special case of vds

  cdo = vgsx - vth - 0.5*(1.0 + fbody) * vdsx;
  cdnorm = cdo * vdsx;
  cdl = beta * cdnorm;
  beta *= fgate;
  Ids = beta * cdnorm;

  // Velocity saturation factor
  if (VMAX != 0.0)
  {
    fdrain = 1.0 / (1.0 + vdsx*onvdsc);
    fd2 = fdrain * fdrain;
    arga = fd2 * vdsx * onvdsc * onfg;
    Ids *= fdrain;
    beta *= fdrain;
  }

  //Channel length modulation
  if (MODE*x[0] <= vdsat) goto line700;

  if (VMAX <= 0.0) goto line510;

  if (alpha == 0.0) goto line700;

  cdsat = Ids;
  gdsat = cdsat * (1.0 - fdrain) * onvdsc;
  if (gdsat < 1.0e-12)
    gdsat = 1.0e-12;

  emax = KAPPA * cdsat / (leff * gdsat);
  arga = 0.5 * emax * alpha;
  argc = KAPPA * alpha;
  argb = sqrt(arga*arga + argc*(MODE * x[0] - vdsat));
  delxl = argb - arga;
  goto line520;

	line510:
	delxl = sqrt(KAPPA * alpha * (MODE*x[0]-vdsat));

	//punch through approximation
	line520:
	if (delxl > (0.5 * leff))
	{
		delxl = leff - (leff*leff/(4.0*delxl));
	}

	//saturation region
	dlonxl = delxl / leff;
	xlfact = 1.0 / (1.0 - dlonxl);
	Ids *= xlfact;

	//weak inversion
	line700:
	if (T1 < von)
	{
		onxn = 1.0 / xn;
		ondvt = onxn / vt;
		wfact = exp((T1 - von) * ondvt);
		Ids *= wfact;
	}
  goto the_end;

	line900:
	Ids = 0.0;
	beta *= fgate;

	the_end: //the end of dc analysis

	von *= TYPE;
	vdsat *=TYPE;

  //Capacitor calculations begin here
  // The Yang-Chatterjee charge model has been used

	adouble qdrn, qsrc, qbulk, qgate, vfb;
	double Co = cox * weff * leff;

	vfb = VT0 - GAMMA*sqrt(PHI) - PHI;

	if (x[1] <= (vfb + x[2])) // accumulation region
	{
		qdrn = 0.0;
		qsrc = 0.0;
		qbulk = -Co * (x[1] - vfb - x[2]);
		qgate = -(qdrn + qsrc + qbulk);
	}
	else if ((x[1] > (vfb + x[2])) && (x[1] <= vth)) //cutoff
	{
		qdrn = 0.0;
		qsrc = 0.0;
		qbulk = Co/2.0 * pow(GAMMA,2.0) - Co/2.0 * GAMMA * sqrt(pow(GAMMA,2.0) + 4.0 * (x[1] - vfb - x[2]));
		qgate = -(qdrn + qsrc + qbulk);
	}
	else if ((x[1] > vth) && (x[1] <= (x[0]+vth))) //saturation
	{
		qdrn = 0.0;
		qsrc = -2.0/3.0 * Co * (x[1] - vth);
		qbulk = Co * (vfb + PHI - vth);
		qgate = -(qdrn + qsrc + qbulk);
	}
	else if (x[1] > (x[0] + vth)) //linear
	{
		T1 = pow(x[0],2.0) / (8.0 * (x[1] - vth - 0.5*x[0]));
		T2 = 0.5 * (x[1] - vth);
		qdrn = - Co * (T1 + T2 - 0.75*x[0]);
		qsrc = -Co * (T2 - T1/3.0 + 0.25*x[0]);
		qbulk = Co/2.0 * pow(GAMMA,2.0) - Co/2.0 * GAMMA * sqrt(pow(GAMMA,2.0) + 4.0 * (x[1] - vfb - x[2]));
		qgate = -(qdrn + qsrc + qbulk);
	}

  y1[0] = MODE * qdrn;
  y1[1] = MODE * qgate;
  y1[2] = MODE * qsrc;

  z1[0] = MODE * Ids; //DC Drain current
  z1[1] = 0.0; //DC Gate current
  z1[2] = MODE * -Ids; //DC Source current

  z1[3] = x[0] - x[2]; //Vdb
  z1[4] = x[1] - x[2]; //Vgb
  z1[5] = -x[2];//Vsb
}

void Mosp3::eval2(adoublev& dy1, adoublev& z1, adoublev& vp, adoublev& ip)
{
  vp[0] = z1[3]; //Vdb
  vp[1] = z1[4]; //Vgb
  vp[2] = z1[5]; //Vsb

  ip[0] = z1[0] + dy1[0]; //Drain current  Drain charge derivative
  ip[1] = z1[1] + dy1[1]; //Gate current, gate charge derivative
  ip[2] = z1[2] - dy1[2]; //Source current, source charge derivative
}

