// Modified for fREEDA 1.3 compliance FPH 1/7/08
#include "../../../../network/ElementManager.h"
#include "../../../../analysis/FreqMNAM.h"
#include "../../../../analysis/TimeMNAM.h"
#include "VpulseAg.h"
#include "../../../../analysis/TimeDomainSV.h"

// Static members
const unsigned VpulseAg::n_par = 15; 

// Element information
ItemInfo VpulseAg::einfo = {
  "vpulseag",
  "Pulsed envelope linear FM-chirp voltage source approximating Agilent PSG behavior",
  "Frank P. Hart/Sonali Luniya"
  DEFAULT_ADDRESS"elements/VpulseAg.html",
  "2005_01_05"
};

// Parameter information
ParmInfo VpulseAg::pinfo[] = {
  {"v1", "Initial value (V)", TR_DOUBLE, false},
  {"v2", "Pulsed value (V)", TR_DOUBLE, false},
  {"td", "Delay time (s)", TR_DOUBLE, false},
  {"tr", "Rise time (s)", TR_DOUBLE, false}, // 10% to 90%
  {"tf", "Fall time (s)", TR_DOUBLE, false}, // 90% to 10%
  {"pw", "Pulse width (s)", TR_DOUBLE, false},
  {"per", "Period (s)", TR_DOUBLE, false},
  {"ec", "Envelope Character ()", TR_INT, false}, //0-trap, 1-exp
  {"fo", "Center Frequency (Hz)", TR_DOUBLE, false},
  {"deltaf", "Sweep Frequency Range (Hz)", TR_DOUBLE, false},
  {"chirpdir", "Chirp Direction ()", TR_INT, false},
  {"phi", "Fixed Phase (degrees)", TR_DOUBLE, false},
  {"null_rise", "Nullify rise", TR_BOOLEAN, false},
  {"null_chirp", "Nullify chirp", TR_BOOLEAN, false},
  {"null_fall", "Nullify fall", TR_BOOLEAN, false}
};


VpulseAg::VpulseAg(const string& iname) : Element(&einfo, pinfo, n_par, iname)
{

  // Set default parameter values
  paramvalue[0] = &(v1 = zero);
  paramvalue[1] = &(v2 = zero);
  paramvalue[2] = &(td = zero);
  paramvalue[3] = &(tr = zero);
  paramvalue[4] = &(tf = zero);
  paramvalue[5] = &(pw = zero); // also functions as chirp duration
  paramvalue[6] = &(per = zero);
  paramvalue[7] = &(ec = 0); // trapezoidal envelope by default
  paramvalue[8] = &(fo = zero); // if fo = zero, get Vpulsexp beh
  paramvalue[9] = &(deltaf = zero);
  paramvalue[10] = &(chirpdir = 1);
  paramvalue[11] = &(phi = zero);
  paramvalue[12] = &(null_rise = false); // must explicitly specify
  paramvalue[13] = &(null_chirp = false);// on netlist to use this
  paramvalue[14] = &(null_fall = false); // capability

  // Set the number of terminals
  setNumTerms(2);

  // Set flags
  setFlags(LINEAR | ONE_REF | TR_TIME_DOMAIN | SOURCE);
  // Set number of states
  setNumberOfStates(1);

  my_row = 0;
  // int_per_time = zero;
  int_per_time = 0.0; // for accuracy, this should be 0, not 'zero'
  do_source_init = true;

}

unsigned VpulseAg::getExtraRC(const unsigned& eqn_number, const MNAMType& type)
{
  // Keep the equation number assigned to this element
  my_row = eqn_number;
  // Add one extra RC
  return 1;
}

void VpulseAg::getExtraRC(unsigned& first_eqn, unsigned& n_rows) const
{
  assert(my_row);
  first_eqn = my_row;
  n_rows = 1;
}

void VpulseAg::fillMNAM(FreqMNAM* mnam)
{
  assert(my_row);
  // Ask my terminals the row numbers
  mnam->setOnes(getTerminal(0)->getRC(), getTerminal(1)->getRC(), my_row);

  // This element behaves as a short circuit for AC/HB analysis.
  // Since the source vector is assumed to be initialized to zero,
  // we do not need to fill it if freq != frequency.
  return;
}


void VpulseAg::fillMNAM(TimeMNAM* mnam)
{
  assert(my_row);
  // Ask my terminals the row numbers
  mnam->setMOnes(getTerminal(0)->getRC(), getTerminal(1)->getRC(), my_row);
  return;
}


void VpulseAg::fillSourceV(TimeMNAM* mnam)
{
  if (do_source_init) {

     do_source_init = false; // execute only once.
     vn = v2 - v1;	// envelope voltage. 
     phir = phi*pi/180.0; 	// Constant phase

     bnd_tr = true;	// Set to permit one execution of 
     bnd_tf = true;        // if statements at chirp boundaries

     if (chirpdir == 1) { // upchirp
        fchrp_tr = -deltaf/2.0;
        fchrp_tf = deltaf/2.0;
     }
     else { // downchirp
        fchrp_tr = deltaf/2.0;
        fchrp_tf = -deltaf/2.0;
     }
     if (ec == 1) { // exponential envelope
        tau_r = tr/(2.0*log(5.0));
        tau_f = tf/(2.0*log(5.0));
     }
     else { // ec == 0, trapezoidal envelope
         tau_r = tr;
         tau_f = tf;
     }
     pwh = pw;
     mu = deltaf/pw;
     if (per <= zero) { // per not specified, so one-shot
        if (tau_r > tau_f) 
           pwl = 14.0*tau_r;
        else 
           pwl = 14.0*tau_f;
     } 
     else
        pwl = per - tf - pw - tr; // per specified, pulse width low.

  }

  const double& ctime = mnam->getTime();
  double reltime = ctime - td - int_per_time;  // Added td FPH 8/26/03
  if (reltime > per) {
    int_per_time = int((ctime - td)/per) * per;
    reltime = ctime - td - int_per_time;
  }
  double e = zero; 	// total value of source entered in src vector
  double dcv = zero;	// DC value of signal
  double env = zero; 	// Envelope of AC signal
  double acv = zero;	// AC signal
  double fchirp = zero; // instantaneous chirp freq offset from fo
  double freq = zero;   // instantaneous freq.
  double chtime = zero; // reltime for use during chirp; ranges from
                        // -pw/2 to pw/2
  if (int_per_time == zero && ctime < td)
    e = v1;
  // past startup delay and into reltime...
  else { // first get envelope value...
     if (ec == 1) { // exponential envelope
       // function 1 - increasing with + 2nd deriv.
       if (reltime < pwl/2.0 + tr/2.0) {
         dcv = v1;
         env = 0.5*vn*exp((reltime - pwl/2.0 - tr/2.0)/tau_r);
       }

       // function 2 - increasing with - 2nd deriv.
       else if (reltime < pwl/2.0 + tr + pwh/2.0) {
         dcv = v1;
         env = 0.5*vn +
	     0.5*vn*(1-exp((-reltime + pwl/2.0 + tr/2.0)/tau_r));
       }

       // function 3 - decreasing with - 2nd deriv.
       else if (reltime < pwl/2.0 + tr + pwh + tf/2.0) {
         dcv = v1;
         env = 0.5*vn +
	     0.5*vn*(1-exp((reltime -pwl/2.0 -tr -pwh -tf/2.0)/tau_f));

       }

       // function 4 - decreasing with + 2nd deriv.
       else if (reltime < pwl + tr + pwh +tf) {
         dcv = v1;
         env = 0.5*vn*exp((-reltime +pwl/2.0 +tr +pwh +tf/2.0)/tau_f);

       }

       // if here, should be between function 4 and function 1, so v1.
       else {
         dcv = v1;
         env = zero;
       }
       // insert acv vs. time code for exp envelope here ...
       // note that the time splits for acv and env are not the same
       // function 1 - flat with value zero 
       if (reltime < pwl/2.0) {
         dcv = v1; 
         acv = zero; 
       }

       // function 2 - linearly increasing with slope tr
       else if (reltime < pwl/2.0 + tr) {
	 dcv = v1; 
	 freq = fo + fchrp_tr;
	 acv = cos(2.0*pi*freq*reltime + phir);
       }

       // function 3 - flat with value v2
       else if (reltime < pwl/2.0 + tr + pwh) {
	 chtime = reltime - pwl/2.0 - tr - pwh/2.0;
         if (bnd_tr) { // execute only once per pulse
	     bnd_tr = false;
	     if (chirpdir == 1)
                fchirp = (mu/2.0)*chtime;
	     else
                fchirp = -(mu/2.0)*chtime;
	     if (null_rise) phib_tr = 0.0;
	     else phib_tr = 2.0*pi*fo*(reltime-chtime)
		+ 2.0*pi*fchrp_tr*reltime - 2.0*pi*fchirp*chtime;
	 }
         dcv = v1;
	 if (chirpdir == 1) // upchirp
	     fchirp = (mu/2.0)*chtime;
	 else // downchirp
	     fchirp = -(mu/2.0)*chtime;
	 freq = fo + fchirp;
	 acv = cos(2.0*pi*freq*chtime + phib_tr + phir);
       }

       // function 4 - linearly decreasing with slope tf
       else if (reltime < pwl/2.0 + tr + pwh + tf) {
         if (bnd_tf) { // execute only once per pulse
	     bnd_tf = false;
	     chtime = reltime - pwl/2.0 - tr - pwh/2.0;
	     if (chirpdir == 1)
                fchirp = (mu/2.0)*chtime;
	     else
                fchirp = -(mu/2.0)*chtime;
	     phib_tf = 2.0*pi*fo*(chtime-reltime) + phib_tr
		- 2.0*pi*fchrp_tf*reltime + 2.0*pi*fchirp*chtime;
	 }
         dcv = v1;
	 freq = fo + fchrp_tf;
	 //cout << "reltime = " << reltime << "; freq = " << 
	 //   freq << endl;
	 acv = cos(2.0*pi*freq*reltime + phib_tf + phir);
       }

       // flat with value v1.
       else {
         dcv = v1;
	 acv = zero;
         bnd_tr = true; // set up for next pulse...
         bnd_tf = true; // set up for next pulse...
       }
     }
     else { // ec == 0 trapezoidal envelope
       // function 1 - flat with value zero 
       if (reltime < pwl/2.0) {
         dcv = v1; 
         env = zero; 
         acv = zero; 
       }

       // function 2 - linearly increasing with slope tr
       else if (reltime < pwl/2.0 + tr) {
	 dcv = v1; 
	 if (null_rise) env = zero;
	 else env = vn*(reltime-pwl/2.0)/tr;
	 freq = fo + fchrp_tr;
	 acv = cos(2.0*pi*freq*reltime + phir);
       }

       // function 3 - flat with value v2
       else if (reltime < pwl/2.0 + tr + pwh) {
	 chtime = reltime - pwl/2.0 - tr - pwh/2.0;
         if (bnd_tr) { // execute only once per pulse
	     bnd_tr = false;
	     if (chirpdir == 1)
                fchirp = (mu/2.0)*chtime;
	     else
                fchirp = -(mu/2.0)*chtime;
	     if (null_rise) phib_tr = 0.0;
	     else phib_tr = 2.0*pi*fo*(reltime-chtime)
		+ 2.0*pi*fchrp_tr*reltime - 2.0*pi*fchirp*chtime;
	 }
         dcv = v1;
	 if (null_chirp) env = zero;
	 else env = vn;
	 if (chirpdir == 1) // upchirp
	     fchirp = (mu/2.0)*chtime;
	 else // downchirp
	     fchirp = -(mu/2.0)*chtime;
	 freq = fo + fchirp;
	 acv = cos(2.0*pi*freq*chtime + phib_tr + phir);
       }

       // function 4 - linearly decreasing with slope tf
       else if (reltime < pwl/2.0 + tr + pwh + tf) {
         if (bnd_tf) { // execute only once per pulse
	     bnd_tf = false;
	     chtime = reltime - pwl/2.0 - tr - pwh/2.0;
	     if (chirpdir == 1)
                fchirp = (mu/2.0)*chtime;
	     else
                fchirp = -(mu/2.0)*chtime;
	     phib_tf = 2.0*pi*fo*(chtime-reltime) + phib_tr
		- 2.0*pi*fchrp_tf*reltime + 2.0*pi*fchirp*chtime;
	 }
         dcv = v1;
	 if (null_fall) env = zero;
	 else env = vn*(pwl/2.0 +tr +pwh +tf -reltime)/tf;
	 freq = fo + fchrp_tf;
	 //cout << "reltime = " << reltime << "; freq = " << 
	 //   freq << endl;
	 acv = cos(2.0*pi*freq*reltime + phib_tf + phir);
       }

       // flat with value v1.
       else {
         dcv = v1;
         env = zero;
	 acv = zero;
         bnd_tr = true; // set up for next pulse...
         bnd_tf = true; // set up for next pulse...
       }
     }
     e = dcv + env*acv;
  }
  // cout << "ctime = " << ctime << ";   e = " << e << endl;
  mnam->setSource(my_row, e);
  return;
}


void VpulseAg::svTran(TimeDomainSV* tdsv)
{
  if (do_source_init) {

     do_source_init = false; // execute only once.
     vn = v2 - v1;	// envelope voltage. 
     phir = phi*pi/180.0; 	// Constant phase

     bnd_tr = true;	// Set to permit one execution of 
     bnd_tf = true;        // if statements at chirp boundaries

     if (chirpdir == 1) { // upchirp
        fchrp_tr = -deltaf/2.0;
        fchrp_tf = deltaf/2.0;
     }
     else { // downchirp
        fchrp_tr = deltaf/2.0;
        fchrp_tf = -deltaf/2.0;
     }
     if (ec == 1) { // exponential envelope
        tau_r = tr/(2.0*log(5.0));
        tau_f = tf/(2.0*log(5.0));
     }
     else { // ec == 0, trapezoidal envelope
         tau_r = tr;
         tau_f = tf;
     }
     pwh = pw;
     mu = deltaf/pw;
     if (per <= zero) { // per not specified, so one-shot
        if (tau_r > tau_f) 
           pwl = 14.0*tau_r;
        else 
           pwl = 14.0*tau_f;
     } 
     else
        pwl = per - tf - pw - tr; // per specified, pulse width low.

  }
  // Calculate voltage
  double& e = tdsv->u(0);

  if (tdsv->DC())
    e = v1;
  else {

    const double& ctime = tdsv->getCurrentTime();
    double reltime = ctime - td - int_per_time;
    if (reltime > per) {
      int_per_time = int((ctime-td)/per) * per;
      reltime = ctime - td - int_per_time;
    }
    double e = zero; 	// total value of source entered in src vector
    double dcv = zero;	// DC value of signal
    double env = zero; 	// Envelope of AC signal
    double acv = zero;	// AC signal
    double fchirp = zero; // instantaneous chirp freq offset from fo
    double freq = zero;   // instantaneous freq.
    double chtime = zero; // reltime for use during chirp; ranges from
                        // -pw/2 to pw/2
    if (int_per_time == zero && ctime < td)
      e = v1;
    // past startup delay and into reltime...
    else { // first get envelope value...
       if (ec == 1) { // exponential envelope
         // function 1 - increasing with + 2nd deriv.
         if (reltime < pwl/2.0 + tr/2.0) {
           dcv = v1;
           env = 0.5*vn*exp((reltime - pwl/2.0 - tr/2.0)/tau_r);
         }

         // function 2 - increasing with - 2nd deriv.
         else if (reltime < pwl/2.0 + tr + pwh/2.0) {
           dcv = v1;
           env = 0.5*vn +
	       0.5*vn*(1-exp((-reltime + pwl/2.0 + tr/2.0)/tau_r));
         }

         // function 3 - decreasing with - 2nd deriv.
         else if (reltime < pwl/2.0 + tr + pwh + tf/2.0) {
           dcv = v1;
           env = 0.5*vn +
	       0.5*vn*(1-exp((reltime -pwl/2.0 -tr -pwh -tf/2.0)/tau_f));

         }

         // function 4 - decreasing with + 2nd deriv.
         else if (reltime < pwl + tr + pwh +tf) {
           dcv = v1;
           env = 0.5*vn*exp((-reltime +pwl/2.0 +tr +pwh +tf/2.0)/tau_f);

         }

         // if here, should be between function 4 and function 1, so v1.
         else {
           dcv = v1;
           env = zero;
         }
         // insert acv vs. time code for exp envelope here ...
         // note that the time splits for acv and env are not the same
         // function 1 - flat with value zero 
         if (reltime < pwl/2.0) {
           dcv = v1; 
           acv = zero; 
         }

         // function 2 - linearly increasing with slope tr
         else if (reltime < pwl/2.0 + tr) {
	   dcv = v1; 
	   freq = fo + fchrp_tr;
	   acv = cos(2.0*pi*freq*reltime + phir);
         }

         // function 3 - flat with value v2
         else if (reltime < pwl/2.0 + tr + pwh) {
	   chtime = reltime - pwl/2.0 - tr - pwh/2.0;
           if (bnd_tr) { // execute only once per pulse
	       bnd_tr = false;
	       if (chirpdir == 1)
                  fchirp = (mu/2.0)*chtime;
	       else
                  fchirp = -(mu/2.0)*chtime;
	       if (null_rise) phib_tr = 0.0;
	       else phib_tr = 2.0*pi*fo*(reltime-chtime)
		  + 2.0*pi*fchrp_tr*reltime - 2.0*pi*fchirp*chtime;
	   }
           dcv = v1;
	   if (chirpdir == 1) // upchirp
	       fchirp = (mu/2.0)*chtime;
	   else // downchirp
	       fchirp = -(mu/2.0)*chtime;
	   freq = fo + fchirp;
	   acv = cos(2.0*pi*freq*chtime + phib_tr + phir);
         }

         // function 4 - linearly decreasing with slope tf
         else if (reltime < pwl/2.0 + tr + pwh + tf) {
           if (bnd_tf) { // execute only once per pulse
	       bnd_tf = false;
	       chtime = reltime - pwl/2.0 - tr - pwh/2.0;
	       if (chirpdir == 1)
                  fchirp = (mu/2.0)*chtime;
	       else
                  fchirp = -(mu/2.0)*chtime;
	       phib_tf = 2.0*pi*fo*(chtime-reltime) + phib_tr
		  - 2.0*pi*fchrp_tf*reltime + 2.0*pi*fchirp*chtime;
	   }
           dcv = v1;
	   freq = fo + fchrp_tf;
	   //cout << "reltime = " << reltime << "; freq = " << 
	   //   freq << endl;
	   acv = cos(2.0*pi*freq*reltime + phib_tf + phir);
         }

         // flat with value v1.
         else {
           dcv = v1;
	   acv = zero;
           bnd_tr = true; // set up for next pulse...
           bnd_tf = true; // set up for next pulse...
         }
       }
       else { // ec == 0 trapezoidal envelope
         // function 1 - flat with value zero 
         if (reltime < pwl/2.0) {
           dcv = v1; 
           env = zero; 
           acv = zero; 
         }

         // function 2 - linearly increasing with slope tr
         else if (reltime < pwl/2.0 + tr) {
	   dcv = v1; 
	   if (null_rise) env = zero;
	   else env = vn*(reltime-pwl/2.0)/tr;
	   freq = fo + fchrp_tr;
	   acv = cos(2.0*pi*freq*reltime + phir);
         }

         // function 3 - flat with value v2
         else if (reltime < pwl/2.0 + tr + pwh) {
	   chtime = reltime - pwl/2.0 - tr - pwh/2.0;
           if (bnd_tr) { // execute only once per pulse
	       bnd_tr = false;
	       if (chirpdir == 1)
                  fchirp = (mu/2.0)*chtime;
	       else
                  fchirp = -(mu/2.0)*chtime;
	       if (null_rise) phib_tr = 0.0;
	       else phib_tr = 2.0*pi*fo*(reltime-chtime)
		  + 2.0*pi*fchrp_tr*reltime - 2.0*pi*fchirp*chtime;
	   }
           dcv = v1;
	   if (null_chirp) env = zero;
	   else env = vn;
	   if (chirpdir == 1) // upchirp
	       fchirp = (mu/2.0)*chtime;
	   else // downchirp
	       fchirp = -(mu/2.0)*chtime;
	   freq = fo + fchirp;
	   acv = cos(2.0*pi*freq*chtime + phib_tr + phir);
         }

         // function 4 - linearly decreasing with slope tf
         else if (reltime < pwl/2.0 + tr + pwh + tf) {
           if (bnd_tf) { // execute only once per pulse
	       bnd_tf = false;
	       chtime = reltime - pwl/2.0 - tr - pwh/2.0;
	       if (chirpdir == 1)
                  fchirp = (mu/2.0)*chtime;
	       else
                  fchirp = -(mu/2.0)*chtime;
	       phib_tf = 2.0*pi*fo*(chtime-reltime) + phib_tr
	  	- 2.0*pi*fchrp_tf*reltime + 2.0*pi*fchirp*chtime;
	   }
           dcv = v1;
	   if (null_fall) env = zero;
	   else env = vn*(pwl/2.0 +tr +pwh +tf -reltime)/tf;
	   freq = fo + fchrp_tf;
	   //cout << "reltime = " << reltime << "; freq = " << 
	   //   freq << endl;
	   acv = cos(2.0*pi*freq*reltime + phib_tf + phir);
         }

         // flat with value v1.
         else {
           dcv = v1;
           env = zero;
	   acv = zero;
           bnd_tr = true; // set up for next pulse...
           bnd_tf = true; // set up for next pulse...
         }
       }
       e = dcv + env*acv;
    }
    // cout << "ctime = " << ctime << ";   e = " << e << endl;
  } // close bracket on else for if (tdsv->DC())
  // Scale state variable for numerical stability
  tdsv->i(0) = tdsv->getX(0) * 1e-2;
} // close bracket on member function

void VpulseAg::deriv_svTran(TimeDomainSV* tdsv)
{
  tdsv->getJu()(0,0) = zero;
  tdsv->getJi()(0,0) = 1e-2;
}
