// Modified for fREEDA 1.3 compliance FPH 1/7/08
#include "../../../../network/ElementManager.h"
#include "../../../../analysis/FreqMNAM.h"
#include "../../../../analysis/TimeMNAM.h"
#include "../../../../analysis/TimeDomainSV.h"
#include <cstdlib>      // need for rand(),srand(), and RAND_MAX
#include <ctime>	// need for std::time() function
#include <cmath>	// need for log() function
#include "Vwnoise.h"

// using std::time;

// Static members
const unsigned Vwnoise::n_par = 4;

// Element information
ItemInfo Vwnoise::einfo = {
  "vwnoise",
  "White Noise voltage source",
  "Frank P. Hart"
  DEFAULT_ADDRESS"elements/Vwnoise.html",
  "2004_06_01"
};

// Parameter information
ParmInfo Vwnoise::pinfo[] = {
  {"vo", "Offset value (V)", TR_DOUBLE, false},
  {"td", "Delay time (s)", TR_DOUBLE, false},
  {"pin", "Mean power level of the source (dBm)", TR_DOUBLE, false}, 
  {"rref", "Ref. resistance for mean power (Ohms)", TR_DOUBLE, false}, 
};


Vwnoise::Vwnoise(const string& iname) : Element(&einfo, pinfo, n_par, iname)
{

  // Set default parameter values
  paramvalue[0] = &(vo = zero);		// offset v. defined at t=0
  paramvalue[1] = &(td = zero);		// delay time before noise
  paramvalue[2] = &(pin = zero);	// mean power (dBm) of noise 
  paramvalue[3] = &(rref = 50.0);       // Ref. resistance (50 ohm def)

  // Set the number of terminals
  setNumTerms(2);

  // Set flags
  setFlags(LINEAR | ONE_REF | TR_TIME_DOMAIN | SOURCE);
  // Set number of states
  setNumberOfStates(1);

  my_row = 0;

  srand(std::time( 0 ));		// seed the rand function

  do_source_init = true;

}

unsigned Vwnoise::getExtraRC(const unsigned& eqn_number, const MNAMType& type)
{
  // Keep the equation number assigned to this element
  my_row = eqn_number;
  // Add one extra RC
  return 1;
}

void Vwnoise::getExtraRC(unsigned& first_eqn, unsigned& n_rows) const
{
  assert(my_row);
  first_eqn = my_row;
  n_rows = 1;
}

void Vwnoise::fillMNAM(FreqMNAM* mnam)
{
  assert(my_row);
  // Ask my terminals the row numbers
  mnam->setOnes(getTerminal(0)->getRC(), getTerminal(1)->getRC(), my_row);

  // This element behaves as a short circuit for AC/HB analysis.
  // Since the source vector is assumed to be initialized to zero,
  // we do not need to fill it if freq != frequency.
  return;
}


void Vwnoise::fillMNAM(TimeMNAM* mnam)
{
  assert(my_row);
  // Ask my terminals the row numbers
  mnam->setMOnes(getTerminal(0)->getRC(), getTerminal(1)->getRC(), my_row);
  return;
}


void Vwnoise::fillSourceV(TimeMNAM* mnam)
{
  const double& ctime = mnam->getTime();
  const double FRAND_MAX = static_cast<double>(RAND_MAX);

  double e = zero;
  double uvrv = zero;
  double usrv = zero;
  double gvrv = zero;
  double sa = zero;
  // double ss = zero;
  double va = zero;
  if (do_source_init) {
      do_source_init = false;
      ss = sqrt(2.0*pi)*pow(10.0,((pin-27.0+10.0*log10(rref))/20.0)); // mean r.v.
  }

  if (ctime < td)  	// no noise output till past startup delay
    e = vo;
  else {

    uvrv = rand()/FRAND_MAX;	// gen uniform r.v. & normalize...
    gvrv = sqrt(-2.0*log(1-pow(uvrv,2.0)));// xform u to gaussian r.v.
    //gvrv = sqrt(-2.0*log(1-uvrv));// xform u to gaussian r.v.

    usrv = rand()/FRAND_MAX;	// gen uniform r.v. & normalize... 
    if (usrv < 0.5) 		// set sign of random voltage
      sa = -1.0;
    else
      sa = 1.0;
     
    va = ss*sa*gvrv;	// output voltage = mean*sign*gaussian r.v.

    e = vo + va;

  }
  // cout << "ctime = " << ctime << ";   e = " << e << endl;
  mnam->setSource(my_row, e);
  return;
}


void Vwnoise::svTran(TimeDomainSV* tdsv)
{
  // Calculate voltage
  double& e = tdsv->u(0);

  if (tdsv->DC())
    e = vo;
  else {

    const double& ctime = tdsv->getCurrentTime();
    const double FRAND_MAX = static_cast<double>(RAND_MAX);
    double e = zero;
    double uvrv = zero;
    double usrv = zero;
    double gvrv = zero;
    double sa = zero;
    double ss = zero;
    double va = zero;

    if (ctime < td)  	// no noise output till past startup delay
      e = vo;
    else {
    //  srand(std::time( 0 ));		// seed the rand function

      uvrv = rand()/FRAND_MAX;	// gen uniform r.v. & normalize...
      gvrv = sqrt(-2.0*log(1-pow(uvrv,2.0)));// xform u to gaussian r.v.

      usrv = rand()/FRAND_MAX;	// gen uniform r.v. & normalize... 
      if (usrv < 0.5) 		// set sign of random voltage
        sa = -1.0;
      else
        sa = 1.0;
     
      ss = pow(10.0,((pin-27.0+10.0*log10(rref))/20.0)); // mean out r.v.

      va = ss*sa*gvrv;	// output voltage = mean*sign*gaussian r.v.

      e = vo + va;
    }
  }
  // Scale state variable for numerical stability
  tdsv->i(0) = tdsv->getX(0) * 1e-2;
}

void Vwnoise::deriv_svTran(TimeDomainSV* tdsv)
{
  tdsv->getJu()(0,0) = zero;
  tdsv->getJi()(0,0) = 1e-2;
}
