#include "AdolcElement.h"
#include "../analysis/TimeDomainSV.h"
#include "../analysis/FreqDomainSV.h"
#include "../analysis/WaveletDomainSV.h"
#include "adutilsc.h"

bool AdolcElement::SVWSstatus = false;
int AdolcElement::aux_size = 20;
double* AdolcElement::x = NULL;
double* AdolcElement::f = NULL;
double* AdolcElement::jac_col_sel = NULL;

AdolcElement::AdolcElement(ItemInfo* einfo, ParmInfo* param_desc,
const int& numparms, const string& iname)
: Element(einfo, param_desc, numparms, iname),
my_nstates(0), ni_adolc(0), nd_adolc(0)
{
  // Set to true in createTape()
  taped = false;
  currentTime = 0.0;
}

AdolcElement::~AdolcElement()
{
}

// State-variable related methods
void AdolcElement::allocSVWS()
{
  if (SVWSstatus)
  {
    delete [] jac_col_sel;
    delete [] f;
    delete [] x;
  }
  jac_col_sel = new double[aux_size];
  for (int j=0; j<aux_size; j++)
    jac_col_sel[j] = zero;
  x = new double[aux_size];
  f = new double[aux_size];
  // Set flag
  SVWSstatus = true;
}

void AdolcElement::createTape(const DenseIntVector& var)
{
  DenseIntVector novar;
  DenseDoubleVector nodelay;
  createTape(var, novar, novar, novar, nodelay);
}

void AdolcElement::createTape(const DenseIntVector& var,
const DenseIntVector& dvar)
{
  DenseIntVector novar;
  DenseDoubleVector nodelay;
  createTape(var, dvar, novar, novar, nodelay);
}

void AdolcElement::createTape(const DenseIntVector& var,
const DenseIntVector& dvar,
const DenseIntVector& d2var,
const DenseIntVector& t_var,
const DenseDoubleVector& delay)
{
  my_nstates = getNumberOfStates();
  assert(my_nstates);
  // Check sizes
  assert(var.size() <= my_nstates);
  assert(dvar.size() <= my_nstates);
  assert(d2var.size() <= my_nstates);
  assert(t_var.size() <= my_nstates);
  assert(t_var.size() == delay.size());

	this->var.resize(var.size());
	mtl::copy(var, this->var);
	this->dvar.resize(dvar.size());
	mtl::copy(dvar, this->dvar);
	this->d2var.resize(d2var.size());
	mtl::copy(d2var, this->d2var);
	this->t_var.resize(t_var.size());
	mtl::copy(t_var, this->t_var);
	this->delay.resize(delay.size());
	mtl::copy(delay, this->delay);

  // set the number of variables used by adol-c
  ni_adolc = var.size() + dvar.size() + d2var.size() + t_var.size();
  nd_adolc = 2 * my_nstates;
  d2const = var.size() + dvar.size();
  delconst = var.size() + dvar.size() + d2var.size();

  // Find the greater
  int max_dim = nd_adolc;
  if (ni_adolc > nd_adolc)
    max_dim = ni_adolc;
  // Reserve memory if needed
  if (!SVWSstatus || (max_dim > aux_size))
  {
    aux_size = max_dim;
    // realloc space
    allocSVWS();
  }

  // Tape the eval() function
  for (int i=0; i<ni_adolc; i++)
    x[i] = zero;

  // Here we could check if we need retaping
  trace_on(getId());
  // inputs
  adoublev ax(ni_adolc);
  // outputs
  adoublev vp(my_nstates), ip(my_nstates);
  ax <<= x;

  eval(ax, vp, ip);
  vp >>= f;
  ip >>= (f + my_nstates);
  trace_off();
  taped = true;
}

void AdolcElement::reTape()
{
  // Here we could check if we need retaping
  trace_on(getId());
  // inputs
  adoublev ax(ni_adolc);
  // outputs
  adoublev vp(my_nstates), ip(my_nstates);
  ax <<= x;

  eval(ax, vp, ip);
  vp >>= f;
  ip >>= f + my_nstates;
  trace_off();
}

void AdolcElement::svHB_timeX(FreqDomainSV*& fdsv)
{
  // Tell fdsv to prepare the time-domain vectors
  for (unsigned i=0; i<var.size(); i++)
  {
    fdsv->transformX(var[i]);
  }
  for (unsigned i=0; i<dvar.size(); i++)
  {
    fdsv->evaldX_dt(dvar[i]);
    fdsv->transformdX_dt(dvar[i]);
  }
  for (unsigned i=0; i<d2var.size(); i++)
  {
    fdsv->evald2X_dt2(d2var[i]);
    fdsv->transformd2X_dt2(d2var[i]);
  }
  for (unsigned i=0; i<t_var.size(); i++)
  {
    fdsv->evalDelX(t_var[i], delay[i]);
    fdsv->transformDelX(t_var[i]);
  }
}

void AdolcElement::svHB_getx(FreqDomainSV*& fdsv, const int& idx)
{
  // Prepare x
  for (unsigned i=0; i<var.size(); i++)
  {
    x[i] = fdsv->getx(var[i])[idx];
  }
  for (unsigned i=0; i<dvar.size(); i++)
  {
    x[var.size() + i] = fdsv->getdx_dt(dvar[i])[idx];
  }
  for (unsigned i=0; i<d2var.size(); i++)
  {
    x[d2const + i] = fdsv->getd2x_dt2(d2var[i])[idx];
  }
  for (unsigned i=0; i<t_var.size(); i++)
  {
    x[delconst + i] = fdsv->getDelx(t_var[i])[idx];
  }
}

void AdolcElement::svHB(FreqDomainSV* fdsv)
{
  assert(taped);
  const int& noSamples = fdsv->noSamples();

  // Prepare time-domain inputs
  svHB_timeX(fdsv);

  // Loop through all time samples
  for (int idx=0; idx < noSamples ; idx++)
  {
    // prepare x
    svHB_getx(fdsv, idx);

    if(function(getId(), nd_adolc, ni_adolc, x, f) < 0)
			reTape();

		// store output
    for (unsigned i=0; i<my_nstates; i++)
    {
      fdsv->vp(i)[idx] = f[i];
      fdsv->ip(i)[idx] = f[my_nstates + i];
    }
  }

  // Tell fdsv to convert the output back to the freq. domain
  for (unsigned i=0; i<my_nstates; i++)
  {
    fdsv->transformvp(i);
    fdsv->transformip(i);
  }
}

void AdolcElement::deriv_svHB(FreqDomainSV* fdsv)
{
  assert(taped);
  const int& noSamples = fdsv->noSamples();

  // Prepare time-domain inputs
  svHB_timeX(fdsv);

  // Loop through all time samples
  for (int idx=0; idx < noSamples ; idx++)
	{
		// prepare x
		svHB_getx(fdsv, idx);

		// store output
		for (unsigned i=0; i<var.size(); i++)
		{
	   	jac_col_sel[i] = one;
	   	if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
			{
				reTape();
				jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
	   	}
	   	jac_col_sel[i] = zero;
	   	for (unsigned j=0; j<my_nstates; j++)
			{
				// df/dx
				fdsv->Jvp(j,var[i])[idx] = f[j];
				fdsv->Jip(j,var[i])[idx] = f[my_nstates + j];
	   	}
		}
		for (unsigned i=0; i<dvar.size(); i++)
		{
	   	jac_col_sel[var.size() + i] = one;
	   	if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
			{
				reTape();
				jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
	   	}
	   	jac_col_sel[var.size() + i] = zero;
	   	for (unsigned j=0; j<my_nstates; j++)
			{
				// df/d(dx/dt)
				fdsv->Jvpdx_dt(j,dvar[i])[idx] = f[j];
				fdsv->Jipdx_dt(j,dvar[i])[idx] = f[my_nstates + j];
	   	}
		}
		for (unsigned i=0; i<d2var.size(); i++)
		{
	   	jac_col_sel[d2const + i] = one;
	   	if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
			{
				reTape();
				jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
	   	}
	   	jac_col_sel[d2const + i] = zero;
	   	for (unsigned j=0; j<my_nstates; j++)
			{
				// df/d(d2x/dt2)
				fdsv->Jvpd2x_dt2(j,d2var[i])[idx] = f[j];
				fdsv->Jipd2x_dt2(j,d2var[i])[idx] = f[my_nstates + j];
	   	}
		}
		for (unsigned i=0; i<t_var.size(); i++)
		{
	   	jac_col_sel[delconst + i] = one;
	   	if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
			{
				reTape();
				jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
	   	}
	   	jac_col_sel[delconst + i] = zero;
	   	for (unsigned j=0; j<my_nstates; j++)
			{
				// df/d(x-tau)
				fdsv->JvpDelx(j,t_var[i])[idx] = f[j];
				fdsv->JipDelx(j,t_var[i])[idx] = f[my_nstates + j];
	   	}
		}
   }
  // Clear Jacobians
  for (unsigned i=0; i<my_nstates; i++)
	{
		for (unsigned j=0; j<my_nstates; j++)
		{
	   	mtl::set_value(fdsv->getJacVp(i,j), zero);
	   	mtl::set_value(fdsv->getJacIp(i,j), zero);
		}
	}
  //  fdsv->buildJ(my_nstates);
  // Tell fdsv to buid the jacobian in the freq. domain.
  for (unsigned i=0; i<var.size(); i++)
		fdsv->addJ(my_nstates, var[i]);
  for (unsigned i=0; i<dvar.size(); i++)
		fdsv->addJdx_dt(my_nstates, dvar[i]);
  for (unsigned i=0; i<d2var.size(); i++)
		fdsv->addJd2x_dt2(my_nstates, d2var[i]);
  for (unsigned i=0; i<t_var.size(); i++)
		fdsv->addJDelx(my_nstates, t_var[i], delay[i]);
}

void AdolcElement::svTran(TimeDomainSV* tdsv)
{
  assert(taped);

  // Prepare x
  for (unsigned i=0; i<var.size(); i++)
    x[i] = tdsv->getX(var[i]);

  for (unsigned i=0; i<dvar.size(); i++)
    x[var.size() + i] = tdsv->getdX_dt(dvar[i]);

  for (unsigned i=0; i<d2var.size(); i++)
    x[d2const + i] = tdsv->getd2X_dt2(d2var[i]);

  for (unsigned i=0; i<t_var.size(); i++)
    x[delconst + i] = tdsv->getDelayedX(t_var[i], delay[i]);

  //if(function(getId(), nd_adolc, ni_adolc, x, f) < 0)
  reTape();

  // set the current time
  currentTime = tdsv->getCurrentTime();

  for (unsigned i=0; i<my_nstates; i++)
  {
    tdsv->u(i) = f[i];
    tdsv->i(i) = f[my_nstates + i];
  }
  return;
}

void AdolcElement::deriv_svTran(TimeDomainSV* tdsv)
{
  assert(taped);

  // Prepare x
  for (unsigned i=0; i<var.size(); i++)
    x[i] = tdsv->getX(var[i]);

  for (unsigned i=0; i<dvar.size(); i++)
    x[var.size() + i] = tdsv->getdX_dt(dvar[i]);

  for (unsigned i=0; i<d2var.size(); i++)
    x[d2const + i] = tdsv->getd2X_dt2(d2var[i]);

  for (unsigned i=0; i<t_var.size(); i++)
    x[delconst + i] = tdsv->getDelayedX(t_var[i], delay[i]);

  // Evaluate derivatives
  tdsv->cleanJac();

  DoubleDenseColMatrix& Ju = tdsv->getJu();
  DoubleDenseColMatrix& Ji = tdsv->getJi();

  for (unsigned i=0; i<var.size(); i++)
  {
    jac_col_sel[i] = one;
    if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
    {
      reTape();
      jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
    }
    jac_col_sel[i] = zero;
    for (unsigned j=0; j<my_nstates; j++)
    {
      // df/dx
      Ju(j,var[i]) = f[j];
      Ji(j,var[i]) = f[my_nstates + j];
    }
  }

  for (unsigned i=0; i<dvar.size(); i++)
  {
    jac_col_sel[var.size() + i] = tdsv->getdx_dtFactor();
    if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
    {
      reTape();
      jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
    }
    jac_col_sel[var.size() + i] = zero;
    for (unsigned j=0; j<my_nstates; j++)
    {
      // df/d(dx/dt)
      Ju(j,dvar[i]) += f[j];
      Ji(j,dvar[i]) += f[my_nstates + j];
    }
  }

  for (unsigned i=0; i<d2var.size(); i++)
  {
    jac_col_sel[d2const + i] = tdsv->getd2x_dt2Factor();
    if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
    {
      reTape();
      jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
    }
    jac_col_sel[d2const + i] = zero;
    for (unsigned j=0; j<my_nstates; j++)
    {
      // df/d(d2x/dt2)
      Ju(j,d2var[i]) += f[j];
      Ji(j,d2var[i]) += f[my_nstates + j];
    }
  }

  for (unsigned i=0; i<t_var.size(); i++)
  {
    jac_col_sel[delconst + i] = tdsv->getDelayFactor(delay[i]);
    if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
    {
      reTape();
      jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
    }
    jac_col_sel[delconst + i] = zero;
    for (unsigned j=0; j<my_nstates; j++)
    {
      // df/d(x-tau)
      Ju(j,t_var[i]) += f[j];
      Ji(j,t_var[i]) += f[my_nstates + j];
    }
  }

  return;
}

void AdolcElement::svWav(WaveletDomainSV* wdsv)
{
	assert(taped);
	const int& noSamples = wdsv->noSamples();
  //
	// Loop through all time samples
	for (int idx=0; idx < noSamples ; idx++)
	{
		// Prepare x
		for (unsigned i=0; i<var.size(); i++)
		{
			x[i] = wdsv->getx(var[i])[idx];
		}
		for (unsigned i=0; i<dvar.size(); i++)
		{
			x[var.size() + i] = wdsv->getdx_dt(dvar[i])[idx];
		}
		// Leave this until we write the missing code
		assert(d2var.size() == 0);
		// for (unsigned i=0; i<d2var.size(); i++)
		// {
			// x[d2const + i] = wdsv->getd2x_dt2(d2var[i])[idx];
		// }
		// ***************************************************
		// IMPORTANT:
		// delays are not yet implemented for wavelet transient.
		// The element just receives the state variable.
		// ***************************************************
		int delconst = var.size() + dvar.size();
		for (unsigned i=0; i<t_var.size(); i++)
		{
			x[delconst + i] = wdsv->getx(t_var[i])[idx];
		}
		if(function(getId(), nd_adolc, ni_adolc, x, f) < 0)
			reTape();

		// store output
		for (unsigned i=0; i<my_nstates; i++)
		{
			wdsv->vp(i, idx) = f[i];
			wdsv->ip(i, idx) = f[my_nstates + i];
		}
	}
	cout << "------------------------------------" << endl;
}


void AdolcElement::deriv_svWav(WaveletDomainSV* wdsv)
{
	assert(taped);
	const int& noSamples = wdsv->noSamples();

	// Loop through all time samples
	for (int idx=0; idx < noSamples ; idx++)
	{
		// Prepare x
		for (unsigned i=0; i<var.size(); i++)
		{
			x[i] = wdsv->getx(var[i])[idx];
		}
		for (unsigned i=0; i<dvar.size(); i++)
		{
			x[var.size() + i] = wdsv->getdx_dt(dvar[i])[idx];
		}
		// Leave this until we write the missing code
		assert(d2var.size() == 0);
		// for (unsigned i=0; i<d2var.size(); i++)
		// {
			// x[d2const + i] = wdsv->getd2x_dt2(d2var[i])[idx];
		// }
		// ***************************************************
		// IMPORTANT:
		// delays are not yet implemented for wavelet transient.
		// The element just receives the state variable.
		// ***************************************************
		int delconst = var.size() + dvar.size();
		for (unsigned i=0; i<t_var.size(); i++)
		{
			x[delconst + i] = wdsv->getx(t_var[i])[idx];
		}

		// store output
		for (unsigned i=0; i<var.size(); i++)
		{
			jac_col_sel[i] = one;
			if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
			{
				reTape();
				jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
			}
			jac_col_sel[i] = zero;
			for (unsigned j=0; j<my_nstates; j++)
			{
				// df/dx
				wdsv->Jvp(j,var[i])[idx] = f[j];
				wdsv->Jip(j,var[i])[idx] = f[my_nstates + j];
			}
		}
		for (unsigned i=0; i<dvar.size(); i++)
		{
			jac_col_sel[var.size() + i] = one;
			if (jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f) < 0)
			{
				reTape();
				jac_vec(getId(), nd_adolc, ni_adolc, x, jac_col_sel, f);
			}
			jac_col_sel[var.size() + i] = zero;
			for (unsigned j=0; j<my_nstates; j++)
			{
				// df/d(dx/dt)
				wdsv->Jvpdx_dt(j,dvar[i])[idx] = f[j];
				wdsv->Jipdx_dt(j,dvar[i])[idx] = f[my_nstates + j];
			}
		}
		// Write d2/dx^2 and (x-tau) cases later
	}
	// Clear Jacobians
	wdsv->cleanJac();
	// Tell wdsv to buid the jacobian
	for (unsigned i=0; i<var.size(); i++)
	{
		wdsv->addJ(var[i]);
	}
	for (unsigned i=0; i<dvar.size(); i++)
	{
		wdsv->addJdx_dt(dvar[i]);
	}
}

DenseDoubleVector AdolcElement::getDelayVec()
{
  return delay;
}


