// AdolcElement3 is an alternative base class for nonlinear elements
// with secondary state variables and 3 evaluation levels which
// use Adol-C.
//
// by Carlos E. Christoffersen

#ifndef AdolcElement3_h
#define AdolcElement3_h 1

#include "Element.h"
#include "adouble.h"

class TimeDomainSV;
class FreqDomainSV;
class WaveletDomainSV;

class AdolcElement3 : public Element
{
	public:

  AdolcElement3(ItemInfo* einfo, ParmInfo* param_desc, const int& numparms,
	const string& iname);

  virtual ~AdolcElement3();

  // Get the number of secondary state variables
  virtual unsigned getNumberOfSecStates() const;

  // ---------------------------------------
  // Analysis Methods
  // ---------------------------------------

  // State variable HB.
	//    virtual void svHB(FreqDomainSV* fdsv);
	//    virtual void deriv_svHB(FreqDomainSV* fdsv);

  // State variable transient.
  virtual void svTran(TimeDomainSV* tdsv);
  virtual void deriv_svTran(TimeDomainSV* tdsv);

  // State variable wavelet transient
	//    virtual void svWav(WaveletDomainSV* wdsv);
	//    virtual void deriv_svWav(WaveletDomainSV* wdsv);

	protected:

  // Select required input variables and create the ADOL-C tape We
  // provide reduced versions for elements that not require
  // derivatives or only require first-order derivatives.
  void createTape1(const DenseIntVector& var,
	const DenseIntVector& t_var,
	const DenseDoubleVector& delay,
	const int ny1, const int nz1);

  void createTape2(const int ny2, const int nz2);

  void createTape3();

  // This function is useful for seccionally defined models
  void reTape1();
  void reTape2();
  void reTape3();

  // Generic state variable evaluation routine
  virtual void eval1(adoublev& x, adoublev& xt,
	adoublev& y1, adoublev& z1) = 0;

  virtual void eval2(adoublev& dy1, adoublev& z1,
	adoublev& y2, adoublev& z2) = 0;

  virtual void eval3(adoublev& dy2, adoublev& z2,
	adoublev& vp, adoublev& ip) = 0;

	private:

  void mult(DoubleDenseColMatrix& c, DoubleDenseColMatrix& a,
	DoubleDenseColMatrix& b, const int& kmax);

  // Number of states
  unsigned my_nstates;
  DenseIntVector var, t_var;
  // var2 and var3 are used to address the secondary state variables in tdsv
  DenseIntVector var2, var3;
  DenseDoubleVector delay;
  DoubleDenseColMatrix Jz1, Jy1, Jy2, Jz2, Jy2tot, Jz2tot, Jz3;
  // Aux variables to save additions
  int delconst, ny1, nz1, ny2, nz2, nz3, nlevels;
  // Auxiliary arrays for built-in analysis evaluation routines
  static double *x, *f, *jac_col_sel;
  // Default size of allocated memory
  static int aux_size;
  // Flag to signal if the function is taped (ADOL-C tape)
  bool taped;
  int ni_adolc1, nd_adolc1, ni_adolc2, nd_adolc2, ni_adolc3, nd_adolc3;
  // This flag signal if memory is already allocated.
  static bool SVWSstatus;
  static unsigned big_number;
  void allocSVWS();
  // Auxiliary functions for HB methods
	// void svHB_timeX(FreqDomainSV*& fdsv);
	// void svHB_getx(FreqDomainSV*& fdsv, const int& idx);
};

#endif

